/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef QSTBOX_H
#define QSTBOX_H

#include "Window.h"
#include "Button.h"
#include "Label.h"
#include "Spacer.h"
#include "GUIStyle.h"
#include "Widget.h"
#include "VBox.h"
#include "HBox.h"

#include <SDL.h>

#include <iostream>

extern SDL_Surface* screen;

#define	QSTBOX_BUTTON_INVALID	(-1)
#define	QSTBOX_BUTTON1			(1)
#define	QSTBOX_BUTTON2			(2)

/// A simple class for a question box
class QstBox : public Window {
public:

	/**
		This method sets a new text for this question box.
		\param	Text The new text for this question box
	*/
	virtual inline void SetText(std::string Text) {
		textlabel.SetText(Text);
		Resize(std::max(vbox.GetMinimumSize().x,120),vbox.GetMinimumSize().y);
	}

	/**
		Get the text of this question box.
		\return the text of this question box
	*/
	inline std::string GetText() { return textlabel.GetText(); };

    /**
		Sets the text color for this question box.
		\param	textcolor	    the color of the text (-1 = default color)
        \param	textshadowcolor	the color of the shadow of the text (-1 = default color)
	*/
	virtual inline void SetTextColor(int textcolor, int textshadowcolor = -1) {
		textlabel.SetTextColor(textcolor, textshadowcolor);
		button1.SetTextColor(textcolor, textshadowcolor);
		button2.SetTextColor(textcolor, textshadowcolor);
	}

	/**
		This method resized the question box to width and height. This method should only be
		called if the new size is a valid size for this question box (See ResizingXAllowed,
		ResizingYAllowed, GetMinumumSize).
		\param	width	the new width of this question box
		\param	height	the new height of this question box
	*/
	virtual void Resize(Uint32 width, Uint32 height) {
		Window::Resize(width,height);
		Position.x = (screen->w - GetSize().x)/2;
		Position.y = (screen->h - GetSize().y)/2;
	}

	/**
		This method is called by the window widget if it requests a resizing of
		this window.
	*/
	virtual void ResizeAll() {
		// QstBox should get bigger if content changes
		if(pWindowWidget != NULL) {
			Point newSize = pWindowWidget->GetMinimumSize();
			newSize.x = std::max(newSize.x,120);
			newSize.y = std::max(newSize.y,30);
			Resize(newSize.x,newSize.y);
		}
	};

	/**
		The number of the pressed button.
		\return	the pressed button (either QSTBOX_BUTTON_INVALID, QSTBOX_BUTTON1 or QSTBOX_BUTTON2)
	*/
	int GetPressedButtonID() const {
		return pressedButtonID;
	}

	/**
		This static method creates a dynamic question box object with Text as the text in the question box.
		The idea behind this method is to simply create a new question box on the fly and
		add it as a child window of some other window. If the window gets closed it will be freed.
		\param	Text		    The question box text
		\param	Button1Text     Text of button 1
		\param	Button2Text     Text of button 2
		\param  DefaultFocus    Button which gets the focus on showing the question box
		\return	The new question box (will be automatically destroyed when it's closed)
	*/
	static QstBox* Create(std::string Text, std::string Button1Text = "No", std::string Button2Text = "Yes", int DefaultFocus = QSTBOX_BUTTON_INVALID) {
		QstBox* qstbox = new QstBox(Text, Button1Text, Button2Text, DefaultFocus);
		qstbox->pAllocated = true;
		return qstbox;
	}

protected:
	/** protected constructor (See Create)
		\param	Text		    Text of this question box
		\param	Button1Text     Text of button 1
		\param	Button2Text     Text of button 2
		\param  DefaultFocus    Button which gets the focus on showing the question box
	*/
	QstBox(std::string Text, std::string Button1Text, std::string Button2Text, int DefaultFocus)
	 : Window(50,50,50,50), pressedButtonID(QSTBOX_BUTTON_INVALID) {
		init(Text, Button1Text, Button2Text, DefaultFocus);
	}

	/// destructor
	virtual ~QstBox() {
    }

private:
	/**
		Initialization helper method.
		\param	Text		    Text of this question box
		\param	Button1Text	    Text of button 1
		\param	Button2Text     Text of button 2
		\param  DefaultFocus    Button which gets the focus on showing the question box
	*/
	void init(std::string Text, std::string Button1Text, std::string Button2Text, int DefaultFocus) {
		SetWindowWidget(&vbox);
		vbox.AddWidget(VSpacer::Create(6));
		vbox.AddWidget(&textlabel);
		vbox.AddWidget(VSpacer::Create(3));
		vbox.AddWidget(&hbox);
		vbox.AddWidget(VSpacer::Create(6));
		hbox.AddWidget(Spacer::Create(), 0.2);
		hbox.AddWidget(&vbox2, 0.6);
		vbox2.AddWidget(VSpacer::Create(4));
		button1.SetText(Button1Text);
		button1.SetOnClick(std::bind(&QstBox::OnButton, this, QSTBOX_BUTTON1));
		hbox2.AddWidget(&button1);
		hbox2.AddWidget(HSpacer::Create(6));
		button2.SetText(Button2Text);
		button2.SetOnClick(std::bind(&QstBox::OnButton, this, QSTBOX_BUTTON2));
		hbox2.AddWidget(&button2);
		vbox2.AddWidget(&hbox2);
		vbox2.AddWidget(VSpacer::Create(4));
		hbox.AddWidget(Spacer::Create(), 0.2);
		SetText(Text);
		textlabel.SetAlignment(Alignment_HCenter);

		if(DefaultFocus == QSTBOX_BUTTON1) {
            button1.SetActive();
		} else if(DefaultFocus == QSTBOX_BUTTON2) {
            button2.SetActive();
		}
	}

	/**
		This method is called when one of the buttons is pressed.
	*/
	virtual void OnButton(int btnID) {
		pressedButtonID = btnID;

		Window* pParentWindow = dynamic_cast<Window*>(GetParent());
		if(pParentWindow != NULL) {
			pParentWindow->CloseChildWindow();
		}
	}

	VBox vbox;					///< vertical box
	HBox hbox;					///< horizontal box
	VBox vbox2;					///< inner vertical box;
	HBox hbox2;					///< inner horizontal box;
	Label textlabel;			///< label that contains the text
	TextButton button1;			///< button 1
	TextButton button2;			///< button 2
	int	pressedButtonID;		///< the pressed button
};

#endif // QSTBOX_H
