/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <Bullet.h>

#include <globals.h>

#include <FileClasses/GFXManager.h>
#include <SoundPlayer.h>
#include <ObjectBase.h>
#include <Game.h>
#include <Map.h>
#include <House.h>
#include <Explosion.h>

#include <misc/draw_util.h>
#include <misc/strictmath.h>

#include <algorithm>


Bullet::Bullet(Uint32 shooterID, Coord* newRealLocation, Coord* newRealDestination, Uint32 bulletID, int damage, bool air)
{
    airAttack = air;

    this->shooterID = shooterID;

    ObjectBase* pShooter = currentGame->getObjectManager().getObject(shooterID);
    if(pShooter == NULL) {
        owner = NULL;
    } else {
        owner = pShooter->getOwner();
    }

	this->bulletID = bulletID;

    this->damage = damage;

    int inaccuracy = Bullet::init();

    destination.x = newRealDestination->x + currentGame->RandomGen.rand(-inaccuracy, inaccuracy);
	destination.y = newRealDestination->y + currentGame->RandomGen.rand(-inaccuracy, inaccuracy);

	if(bulletID == Bullet_Sonic) {
		int diffX = destination.x - newRealLocation->x;
        int diffY = destination.y - newRealLocation->y;

		float ratio;
		float square_root;

		int weaponrange = currentGame->objectData.data[Unit_SonicTank].weaponrange;

		if((diffX == 0) && (diffY == 0)) {
			diffY = weaponrange*TILESIZE;
		}

		square_root = strictmath::sqrt((float)(diffX*diffX + diffY*diffY));
		ratio = (weaponrange*TILESIZE)/square_root;
		destination.x = newRealLocation->x + (int)(((float)diffX)*ratio);
		destination.y = newRealLocation->y + (int)(((float)diffY)*ratio);
	}

	realX = (float)newRealLocation->x;
	realY = (float)newRealLocation->y;
	source.x = newRealLocation->x;
	source.y = newRealLocation->y;
	location.x = newRealLocation->x/TILESIZE;
	location.y = newRealLocation->y/TILESIZE;

	angle = dest_angle(*newRealLocation, destination);
	drawnAngle = (int)((float)numFrames*angle/256.0);

    xSpeed = speed * strictmath::cos(angle * conv2char);
	ySpeed = speed * -strictmath::sin(angle * conv2char);
}

Bullet::Bullet(InputStream& stream)
{
	bulletID = stream.readUint32();

	airAttack = stream.readBool();
	damage = stream.readSint32();

	shooterID = stream.readUint32();
	Uint32 x = stream.readUint32();
	if(x < MAX_PLAYERS) {
		owner = currentGame->house[x];
	} else {
		owner = currentGame->house[0];
	}

    source.x = stream.readSint32();
	source.y = stream.readSint32();
	destination.x = stream.readSint32();
	destination.y = stream.readSint32();
    location.x = stream.readSint32();
	location.y = stream.readSint32();
	realX = stream.readFloat();
	realY = stream.readFloat();

    xSpeed = stream.readFloat();
	ySpeed = stream.readFloat();

	drawnAngle = stream.readSint8();
    angle = stream.readFloat();

	Bullet::init();
}

int Bullet::init()
{
    groundBlocked = false;
	int inaccuracy;

	int houseID = (owner == NULL) ? HOUSE_HARKONNEN : owner->getHouseID();

	switch(bulletID) {
        case(Bullet_Inf_Gun): {
            damageRadius = 8;
            groundBlocked = true;
            inaccuracy = 8;
            speed = 36.0f;
            numFrames = 1;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_Small, houseID);
        } break;

        case(Bullet_Gun): {
            damageRadius = 8;
            groundBlocked = true;
            inaccuracy = 8;
            speed = 36.0f;
            numFrames = 1;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_Small, houseID);
        } break;

        case(Bullet_DRocket): {
            damageRadius = 64;
            inaccuracy = 64;
            speed = 16.0f;
            numFrames = 16;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_MediumRocket, houseID);
        } break;

        case(Bullet_LargeRocket): {
            damageRadius = 64;
            inaccuracy = 0;
            speed = 16.0f;
            numFrames = 16;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_LargeRocket, houseID);
        } break;

        case(Bullet_Rocket): {
            damageRadius = 16;
            ObjectBase* pShooter = currentGame->getObjectManager().getObject(shooterID);
            if((pShooter != NULL) && (pShooter->getItemID() == Unit_Launcher)) {
                inaccuracy = 64;
            } else {
                inaccuracy = 32;
            }
            speed = 20.0f;
            numFrames = 16;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_MediumRocket, houseID);
        } break;

        case(Bullet_Shell): {
            damageRadius = 12;
            groundBlocked = true;
            inaccuracy = 16;
            speed = 36.0f;
            numFrames = 1;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_Medium, houseID);
        } break;

        case(Bullet_SmallRocket): {
            damageRadius = 20;
            inaccuracy = 24;
            speed = 16.0f;
            numFrames = 16;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_SmallRocket, houseID);
        } break;

        case(Bullet_SmallRocketHeavy): {
            damageRadius = 20;
            inaccuracy = 24;
            speed = 16.0f;
            numFrames = 16;
            graphic = pGFXManager->getObjPic(ObjPic_Bullet_SmallRocket, houseID);
        } break;

        case(Bullet_Sonic): {
            damageRadius = 28;
            inaccuracy = 8;
            speed = 6.0f;
            numFrames = 1;
            SDL_Surface** tmpSurfaceStack = pGFXManager->getObjPic(ObjPic_Bullet_Sonic, houseID);
            graphic = new SDL_Surface*[NUM_ZOOMLEVEL];
            for(int z = 0; z < NUM_ZOOMLEVEL; z++) {
                graphic[z] =  copySurface(tmpSurfaceStack[z]);	//make a copy of the image
            }
        } break;

        case(Bullet_Sandworm): {
            fprintf(stderr,"Bullet::init(): Unknown Bullet_Sandworm not allowed.\n");
            inaccuracy = 0;
            graphic = NULL;
        } break;

        default: {
            fprintf(stderr,"Bullet::init(): Unknown Bullet type %d.\n",bulletID);
            inaccuracy = 0;
            graphic = NULL;
        } break;
	}

	return inaccuracy;
}


Bullet::~Bullet()
{
    if((bulletID == Bullet_Sonic) && (graphic != NULL)) {
        for(int z = 0; z < NUM_ZOOMLEVEL; z++) {
            SDL_FreeSurface(graphic[z]);
        }
        delete [] graphic;
        graphic = NULL;
    }
}

void Bullet::save(OutputStream& stream) const
{
	stream.writeUint32(bulletID);

	stream.writeBool(airAttack);
    stream.writeSint32(damage);

    stream.writeUint32(shooterID);
    stream.writeUint32(owner->getHouseID());

	stream.writeSint32(source.x);
	stream.writeSint32(source.y);
	stream.writeSint32(destination.x);
	stream.writeSint32(destination.y);
    stream.writeSint32(location.x);
	stream.writeSint32(location.y);
	stream.writeFloat(realX);
	stream.writeFloat(realY);

    stream.writeFloat(xSpeed);
	stream.writeFloat(ySpeed);

	stream.writeSint8(drawnAngle);
    stream.writeFloat(angle);
}


void Bullet::blitToScreen()
{
    int imageW = graphic[currentZoomlevel]->w/numFrames;
    int imageH = graphic[currentZoomlevel]->h;

	if(screenborder->isInsideScreen( Coord(lround(realX), lround(realY)), Coord(imageW, imageH)) == false) {
        return;
	}

    SDL_Rect source = { (numFrames > 1) ? drawnAngle * imageW : 0, 0, imageW, imageH };
	SDL_Rect dest = {   screenborder->world2screenX(realX) - imageW/2,
                        screenborder->world2screenY(realY) - imageH/2,
                        imageW, imageH };

    if(bulletID == Bullet_Sonic && !currentGame->isGamePaused()) {
        SDL_Surface** mask = pGFXManager->getObjPic(ObjPic_Bullet_Sonic,(owner == NULL) ? HOUSE_HARKONNEN : owner->getHouseID());

        if(mask[currentZoomlevel]->format->BitsPerPixel == 8) {
            if ((!SDL_MUSTLOCK(screen) || (SDL_LockSurface(screen) == 0))
                && (!SDL_MUSTLOCK(mask[currentZoomlevel]) || (SDL_LockSurface(mask[currentZoomlevel]) == 0))
                && (!SDL_MUSTLOCK(graphic[currentZoomlevel]) || (SDL_LockSurface(graphic[currentZoomlevel]) == 0)))
            {
                unsigned char *maskPixels = (unsigned char*)mask[currentZoomlevel]->pixels;
                unsigned char *screenPixels = (unsigned char*)screen->pixels;
                unsigned char *surfacePixels = (unsigned char*)graphic[currentZoomlevel]->pixels;
                int	maxX = world2zoomedWorld(screenborder->getRight());
                int	maxY = world2zoomedWorld(screenborder->getBottom());

                for(int i = 0; i < dest.w; i++) {
                    for (int j = 0; j < dest.h; j++) {
                        int x;
                        int y;

                        if(maskPixels[i + j*mask[currentZoomlevel]->pitch] == 0) {
                            //not masked, direct copy
                            x = i;
                            y = j;
                        } else {
                            x = i + world2zoomedWorld(12);
                            y = j;
                        }

                        dest.x = std::max(0,std::min( (int) dest.x, maxX-1));
                        dest.y = std::max(0,std::min( (int) dest.y, maxY-1));

                        if(x < 0) {
                            x = 0;
                        } else if (dest.x + x >= maxX) {
                            x = maxX - dest.x - 1;
                        }

                        if(y < 0) {
                            y = 0;
                        } else if (dest.y + y >= screen->h) {
                            x = screen->h - dest.y - 1;
                        }

                        if((dest.x + x >= 0) && (dest.x + x < screen->w) && (dest.y + y >= 0) && (dest.y + y < screen->h)) {
                            surfacePixels[i + j*graphic[currentZoomlevel]->pitch] = screenPixels[dest.x + x + (dest.y + y)*screen->pitch];
                        } else {
                            surfacePixels[i + j*graphic[currentZoomlevel]->pitch] = 0;
                        }
                    }
                }

                if (SDL_MUSTLOCK(graphic[currentZoomlevel]))
                    SDL_UnlockSurface(graphic[currentZoomlevel]);

                if (SDL_MUSTLOCK(mask[currentZoomlevel]))
                    SDL_UnlockSurface(mask[currentZoomlevel]);

                if (SDL_MUSTLOCK(screen))
                    SDL_UnlockSurface(screen);
            }
        }
    }//end of if sonic

    SDL_BlitSurface(graphic[currentZoomlevel], &source, screen, &dest);
}


void Bullet::update()
{
	//Coord oldLocation = location;

	realX += xSpeed;  //keep the bullet moving by its current speeds
	realY += ySpeed;
	location.x = (int)(realX/TILESIZE);
	location.y = (int)(realY/TILESIZE);

	if((location.x < -5) || (location.x >= currentGameMap->sizeX + 5) || (location.y < -5) || (location.y >= currentGameMap->sizeY + 5)) {
        // it's off the screen => delete it
        bulletList.remove(this);
        delete this;
	} else {
		if(distance_from(source.x, source.y, realX, realY) >= distance_from(source, destination))	{
			realX = destination.x;
			realY = destination.y;
			destroy();
		} else if(bulletID == Bullet_Sonic) {
			//if((location.x != oldLocation.x) || (location.y != oldLocation.y)) {
				Coord realPos = Coord(lround(realX), lround(realY));
				currentGameMap->damage(shooterID, owner, realPos, bulletID,
                                        (int) (((float)damage/2.0f)/((float)currentGame->objectData.data[Unit_SonicTank].weaponrange)),
                                        damageRadius, false);

                realX += xSpeed;  //keep the bullet moving by its current speeds
                realY += ySpeed;

				realPos = Coord(lround(realX), lround(realY));
				currentGameMap->damage(shooterID, owner, realPos, bulletID,
                                        (int) (((float)damage/2.0f)/((float)currentGame->objectData.data[Unit_SonicTank].weaponrange)),
                                        damageRadius, false);
			//}
		} else if( currentGameMap->cellExists(location)
                    && currentGameMap->getCell(location)->hasAGroundObject()
                    && currentGameMap->getCell(location)->getGroundObject()->isAStructure()) {
			if(groundBlocked) {
				destroy();
			}
		}
	}
}


void Bullet::destroy()
{
    Coord position = Coord(lround(realX), lround(realY));

    int houseID = (owner == NULL) ? HOUSE_HARKONNEN : owner->getHouseID();

    switch(bulletID) {
        case Bullet_Inf_Gun:
        case Bullet_Gun:
        {
            currentGameMap->damage(shooterID, owner, position, bulletID, damage, damageRadius, airAttack);
        } break;

        case Bullet_DRocket: {
            currentGameMap->damage(shooterID, owner, position, bulletID, damage, damageRadius, airAttack);
            soundPlayer->playSound(Sound_ExplosionGas);
            currentGame->getExplosionList().push_back(new Explosion(Explosion_Gas,position,houseID));
        } break;

        case Bullet_LargeRocket: {
            soundPlayer->playSound(Sound_ExplosionLarge);

            for(int i = 0; i < 5; i++) {
                for(int j = 0; j < 5; j++) {
                    if (((i != 0) && (i != 4)) || ((j != 0) && (j != 4))) {
                        position.x = lround(realX) + (i - 2)*TILESIZE;
                        position.y = lround(realY) + (j - 2)*TILESIZE;

                        currentGameMap->damage(shooterID, owner, position, bulletID, damage, damageRadius, airAttack);

                        Uint32 explosionID = currentGame->RandomGen.getRandOf(2,Explosion_Large1,Explosion_Large2);
                        currentGame->getExplosionList().push_back(new Explosion(explosionID,position,houseID));
                        screenborder->shakeScreen(22);
                    }
                }
            }
        } break;

        case Bullet_Rocket:
        case Bullet_SmallRocket:
        case Bullet_SmallRocketHeavy:
        {
            currentGameMap->damage(shooterID, owner, position, bulletID, damage, damageRadius, airAttack);
            currentGame->getExplosionList().push_back(new Explosion(Explosion_Small,position,houseID));
        } break;

        case Bullet_Shell: {
            currentGameMap->damage(shooterID, owner, position, bulletID, damage, damageRadius, airAttack);
            currentGame->getExplosionList().push_back(new Explosion(Explosion_Shell,position,houseID));
        } break;

        case Bullet_Sonic:
        case Bullet_Sandworm:
        default: {
            // do nothing
        } break;
    }

    bulletList.remove(this);
    delete this;
}

