/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <Command.h>

#include <globals.h>

#include <Game.h>
#include <House.h>

#include <units/UnitBase.h>
#include <units/Carryall.h>
#include <units/Devastator.h>
#include <units/MCV.h>
#include <units/InfantryBase.h>
#include <structures/BuilderBase.h>
#include <structures/TurretBase.h>
#include <structures/Palace.h>
#include <structures/StarPort.h>

#include <stdexcept>

Command::Command(Uint8 playerID, CMDTYPE id) {
	CommandID = id;
}

Command::Command(Uint8 playerID, CMDTYPE id, Uint32 parameter1) {
	this->playerID = playerID;
	CommandID = id;
	Parameter.push_back(parameter1);
}

Command::Command(Uint8 playerID, CMDTYPE id, Uint32 parameter1, Uint32 parameter2) {
	this->playerID = playerID;
	CommandID = id;
	Parameter.push_back(parameter1);
	Parameter.push_back(parameter2);
}

Command::Command(Uint8 playerID, CMDTYPE id, Uint32 parameter1, Uint32 parameter2, Uint32 parameter3) {
	this->playerID = playerID;
	CommandID = id;
	Parameter.push_back(parameter1);
	Parameter.push_back(parameter2);
	Parameter.push_back(parameter3);
}

Command::Command(Uint8 playerID, CMDTYPE id, Uint32 parameter1, Uint32 parameter2, Uint32 parameter3, Uint32 parameter4) {
	this->playerID = playerID;
	CommandID = id;
	Parameter.push_back(parameter1);
	Parameter.push_back(parameter2);
	Parameter.push_back(parameter3);
	Parameter.push_back(parameter4);
}

Command::Command(Uint8 playerID, Uint8* data, Uint32 length) {
	if(length % 4 != 0) {
		throw std::invalid_argument("Command::Command(): Length must be multiple of 4!");
	}

	if(length < 4) {
		throw std::invalid_argument("Command::Command(): Command must be at least 4 bytes long!");
	}

    this->playerID = playerID;
	CommandID = (CMDTYPE) *((Uint32*) data);

	if(CommandID >= CMD_MAX) {
		throw std::invalid_argument("Command::Command(): CommandID unknown!");
	}

	Uint32* pData = (Uint32*) (data+4);
	for(Uint32 i=0;i<(length-4)/4;i++) {
		Parameter.push_back(*pData);
		pData++;
	}
}

Command::Command(InputStream& stream) {
    playerID = stream.readUint8();
	CommandID = (CMDTYPE) stream.readUint32();
	Parameter = stream.readUint32Vector();
}

Command::~Command() {
}

void Command::save(OutputStream& stream) const {
    stream.writeUint8(playerID);
	stream.writeUint32((Uint32) CommandID);
	stream.writeUint32Vector(Parameter);
	stream.flush();
}

void Command::executeCommand() const {
	switch(CommandID) {

		case CMD_PLACE_STRUCTURE: {
			if(Parameter.size() != 4) {
				throw std::invalid_argument("Command::executeCommand(): CMD_PLACE_STRUCTURE needs 4 Parameters!");
			}
			BuilderBase* pBuilder = dynamic_cast<BuilderBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pBuilder == NULL) {
				return;
			}
			pBuilder->getOwner()->placeStructure(Parameter[0], (int) Parameter[1], (int) Parameter[2], (int) Parameter[3]);
		} break;


		case CMD_UNIT_MOVE2POS: {
			if(Parameter.size() != 4) {
				throw std::invalid_argument("Command::executeCommand(): CMD_UNIT_MOVE2POS needs 4 Parameters!");
			}
			UnitBase* unit = dynamic_cast<UnitBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(unit == NULL) {
                return;
			}
            unit->DoMove2Pos((int) Parameter[1], (int) Parameter[2], (bool) Parameter[3]);
		} break;

		case CMD_UNIT_MOVE2OBJECT: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_UNIT_MOVE2OBJECT needs 2 Parameters!");
			}
			UnitBase* unit = dynamic_cast<UnitBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(unit == NULL) {
                return;
			}
            unit->DoMove2Object((int) Parameter[1]);
		} break;

		case CMD_UNIT_ATTACKPOS: {
			if(Parameter.size() != 4) {
				throw std::invalid_argument("Command::executeCommand(): CMD_UNIT_ATTACKPOS needs 4 Parameters!");
			}
			UnitBase* unit = dynamic_cast<UnitBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(unit == NULL) {
                return;
			}
			unit->DoAttackPos((int) Parameter[1], (int) Parameter[2], (bool) Parameter[3]);
		} break;

		case CMD_UNIT_ATTACKOBJECT: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_UNIT_ATTACKOBJECT needs 2 Parameters!");
			}
			UnitBase* pUnit = dynamic_cast<UnitBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pUnit == NULL) {
                return;
			}
			pUnit->DoAttackObject((int) Parameter[1]);
		} break;

        case CMD_INFANTRY_CAPTURE: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_INFANTRY_CAPTURE needs 2 Parameters!");
			}
			InfantryBase* pInfantry = dynamic_cast<InfantryBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pInfantry == NULL) {
                return;
			}
			pInfantry->DoCaptureObject((int) Parameter[1]);
		} break;

		case CMD_UNIT_SETMODE: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_UNIT_SETMODE needs 2 Parameter!");
			}
			UnitBase* pUnit = dynamic_cast<UnitBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pUnit == NULL) {
                return;
			}
			pUnit->DoSetAttackMode((ATTACKMODE) Parameter[1]);
		} break;

		case CMD_CARRYALL_SETDEPLOYSTRUCTURE: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_CARRYALL_SETDEPLOYSTRUCTURE needs 2 Parameters!");
			}
			Carryall* pCarryall = dynamic_cast<Carryall*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pCarryall == NULL) {
                return;
			}
			pCarryall->DoSetDeployStructure((int) Parameter[1]);
		} break;

		case CMD_DEVASTATOR_STARTDEVASTATE: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_DEVASTATOR_STARTDEVASTATE needs 1 Parameter!");
			}
			Devastator* pDevastator = dynamic_cast<Devastator*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pDevastator == NULL) {
                return;
			}
			pDevastator->DoStartDevastate();
		} break;

		case CMD_MCV_DEPLOY: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_MCV_DEPLOY needs 1 Parameter!");
			}
			MCV* pMCV = dynamic_cast<MCV*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pMCV == NULL) {
                return;
			}
			pMCV->DoDeploy();
		} break;

		case CMD_STRUCTURE_SETDEPLOYPOSITION: {
			if(Parameter.size() != 3) {
				throw std::invalid_argument("Command::executeCommand(): CMD_STRUCTURE_SETDEPLOYPOSITION needs 3 Parameters!");
			}
			StructureBase* pStructure = dynamic_cast<StructureBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pStructure == NULL) {
                return;
			}
			pStructure->DoSetDeployPosition((int) Parameter[1],(int) Parameter[2]);
		} break;

		case CMD_STRUCTURE_REPAIR: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_STRUCTURE_REPAIR needs 1 Parameter!");
			}
			StructureBase* pStructure = dynamic_cast<StructureBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pStructure == NULL) {
                return;
			}
			pStructure->DoRepair();
		} break;

		case CMD_BUILDER_UPGRADE: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_BUILDER_UPGRADE needs 1 Parameter!");
			}
			BuilderBase* pBuilder = dynamic_cast<BuilderBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pBuilder == NULL) {
                return;
			}
			pBuilder->DoUpgrade();
		} break;

		case CMD_BUILDER_PRODUCEITEM: {
			if(Parameter.size() != 3) {
				throw std::invalid_argument("Command::executeCommand(): CMD_BUILDER_PRODUCEITEM needs 3 Parameter!");
			}
			BuilderBase* pBuilder = dynamic_cast<BuilderBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pBuilder == NULL) {
                return;
			}
			pBuilder->DoProduceItem(Parameter[1],(bool) Parameter[2]);
		} break;

		case CMD_BUILDER_CANCELITEM: {
			if(Parameter.size() != 3) {
				throw std::invalid_argument("Command::executeCommand(): CMD_BUILDER_CANCELITEM needs 3 Parameter!");
			}
			BuilderBase* pBuilder = dynamic_cast<BuilderBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pBuilder == NULL) {
                return;
			}
			pBuilder->DoCancelItem(Parameter[1],(bool) Parameter[2]);
		} break;

		case CMD_BUILDER_SETONHOLD: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_BUILDER_SETONHOLD needs 2 Parameters!");
			}
			BuilderBase* pBuilder = dynamic_cast<BuilderBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pBuilder == NULL) {
                return;
			}
			pBuilder->DoSetOnHold((bool) Parameter[1]);
		} break;

		case CMD_PALACE_SPECIALWEAPON: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_PALACE_SPECIALWEAPON needs 1 Parameter!");
			}
			Palace* palace = dynamic_cast<Palace*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(palace == NULL) {
                return;
			}
			palace->DoSpecialWeapon();
		} break;

        case CMD_PALACE_DEATHHAND: {
			if(Parameter.size() != 3) {
				throw std::invalid_argument("Command::executeCommand(): CMD_PALACE_DEATHHAND needs 3 Parameter!");
			}
			Palace* palace = dynamic_cast<Palace*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(palace == NULL) {
                return;
			}
			palace->DoLaunchDeathhand((int) Parameter[1], (int) Parameter[2]);
		} break;

		case CMD_STARPORT_PLACEORDER: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_STARPORT_PLACEORDER needs 1 Parameter!");
			}
			StarPort* pStarport = dynamic_cast<StarPort*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pStarport == NULL) {
                return;
			}
			pStarport->DoPlaceOrder();
		} break;

		case CMD_STARPORT_CANCELORDER: {
			if(Parameter.size() != 1) {
				throw std::invalid_argument("Command::executeCommand(): CMD_STARPORT_CANCELORDER needs 1 Parameter!");
			}
			StarPort* pStarport = dynamic_cast<StarPort*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pStarport == NULL) {
                return;
			}
			pStarport->DoCancelOrder();
		} break;

		case CMD_TURRET_ATTACKOBJECT: {
			if(Parameter.size() != 2) {
				throw std::invalid_argument("Command::executeCommand(): CMD_TURRET_ATTACKOBJECT needs 2 Parameters!");
			}
			TurretBase* pTurret = dynamic_cast<TurretBase*>(currentGame->getObjectManager().getObject(Parameter[0]));
			if(pTurret == NULL) {
                return;
			}
			pTurret->DoAttackObject((int) Parameter[1]);
		} break;

		default: {
			throw std::invalid_argument("Command::executeCommand(): Unknown CommandID!");
		} break;
	}

}

