/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FileClasses/Icnfile.h>
#include <FileClasses/Palette.h>

#include <SDL_endian.h>
#include <stdlib.h>
#include <string.h>
#include <stdexcept>

#define	SIZE_X	16
#define SIZE_Y	16

extern Palette palette;

/// Constructor
/**
	The constructor reads from IcnRWop and MapRWop all data and saves them internal. The SDL_RWops can be readonly but
	must support seeking. Immediately after the Icnfile-Object is constructed both RWops can be closed. All data is saved
	in the class.
	\param	IcnRWop	SDL_RWops to the icn-File. (can be readonly)
	\param	MapRWop	SDL_RWops to the map-File. (can be readonly)
	\param  freesrc A non-zero value means it will automatically close/free the IcnRWop and MapRWop for you.
*/
Icnfile::Icnfile(SDL_RWops* IcnRWop, SDL_RWops* MapRWop, int freesrc)
{
    IcnFiledata = NULL;

	if(IcnRWop == NULL) {
	    if(freesrc && MapRWop != NULL) SDL_RWclose(MapRWop);
		throw std::invalid_argument("Icnfile::Icnfile(): IcnRWop == NULL!");
	} else if(MapRWop == NULL) {
	    if(freesrc) SDL_RWclose(IcnRWop);
		throw std::invalid_argument("Icnfile::Icnfile(): MapRWop == NULL!");
	}

	uint8_t* MapFiledata = NULL;

    try {
        int IcnFilesize = SDL_RWseek(IcnRWop,0,SEEK_END);
        if(IcnFilesize <= 0) {
            throw std::runtime_error("Icnfile::Icnfile(): Cannot determine size of this *.icn-File!");
        }
        IcnFiledata = new uint8_t[IcnFilesize];

        if(SDL_RWseek(IcnRWop,0,SEEK_SET) != 0) {
            throw std::runtime_error("Icnfile::Icnfile(): Seeking in this *.icn-File failed!");
        }

        if(SDL_RWread(IcnRWop, &IcnFiledata[0], IcnFilesize, 1) != 1) {
            throw std::runtime_error("Icnfile::Icnfile(): Reading this *.icn-File failed!");
        }


        int MapFilesize = SDL_RWseek(MapRWop,0,SEEK_END);
        if(MapFilesize <= 0) {
            throw std::runtime_error("Icnfile::Icnfile(): Cannot determine size of this *.map-File!");
        }

        MapFiledata = new uint8_t[MapFilesize];

        if(SDL_RWseek(MapRWop,0,SEEK_SET) != 0) {
            throw std::runtime_error("Icnfile::Icnfile(): Seeking in this *.map-File failed!");
        }

        if(SDL_RWread(MapRWop, &MapFiledata[0], MapFilesize, 1) != 1) {
            throw std::runtime_error("Icnfile::Icnfile(): Reading this *.map-File failed!");
        }

        // now we can start creating the Tilesetindex
        if(MapFilesize < 2) {
            throw std::runtime_error("Icnfile::Icnfile(): This *.map-File is too short!");
        }

        Uint16 NumTilesets = SDL_SwapLE16( *((Uint16 *) MapFiledata));

        if(MapFilesize < NumTilesets * 2) {
            throw std::runtime_error("Icnfile::Icnfile(): This *.map-File is too short!");
        }

        // calculate size for all entries
        Uint16 index = SDL_SwapLE16( ((Uint16*) MapFiledata)[0]);
        for(int i = 1; i < NumTilesets; i++) {
            Uint16 tmp = SDL_SwapLE16( ((Uint16*) MapFiledata)[i]);
            MapfileEntry newMapfileEntry;
            newMapfileEntry.NumTiles = tmp - index;
            Tilesets.push_back(newMapfileEntry);
            index = tmp;
        }
        MapfileEntry newMapfileEntry;
        newMapfileEntry.NumTiles = (MapFilesize/2) - index;
        Tilesets.push_back(newMapfileEntry);

        for(int i = 0; i < NumTilesets; i++) {
            index = SDL_SwapLE16( ((Uint16*) MapFiledata)[i]);

            if((unsigned int) MapFilesize < (index+Tilesets[i].NumTiles)*2 ) {
                throw std::runtime_error("Icnfile::Icnfile(): This *.map-File is too short!");
            }

            // now we can read in
            for(unsigned int j = 0; j < Tilesets[i].NumTiles; j++) {
                Tilesets[i].TileIndices.push_back(SDL_SwapLE16( ((Uint16*) MapFiledata)[index+j]));
            }
        }

        delete [] MapFiledata;
        // reading MAP-File is now finished

        // check if we can access first section in ICN-File
        if(IcnFilesize < 0x20) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: No SSET-Section found!\n");
        }

        SSET = IcnFiledata+0x18;

        // check SSET-Section
        if(		(SSET[0] != 'S')
            ||	(SSET[1] != 'S')
            ||	(SSET[2] != 'E')
            ||	(SSET[3] != 'T')) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: No SSET-Section found!\n");
        }

        SSET_Length = SDL_SwapBE32( *((Uint32*) (SSET + 4))) - 8;
        SSET += 16;

        if(IcnFiledata + IcnFilesize < SSET + SSET_Length) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: SSET-Section is bigger than ICN-File!\n");
        }

        RPAL = SSET + SSET_Length;

        // check RPAL-Section
        if(		(RPAL[0] != 'R')
            ||	(RPAL[1] != 'P')
            ||	(RPAL[2] != 'A')
            ||	(RPAL[3] != 'L')) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: No RPAL-Section found!\n");
        }

        RPAL_Length = SDL_SwapBE32( *((Uint32*) (RPAL + 4)));
        RPAL += 8;

        if(IcnFiledata + IcnFilesize < RPAL + RPAL_Length) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: RPAL-Section is bigger than ICN-File!\n");
        }

        RTBL = RPAL + RPAL_Length;

        // check RTBL-Section
        if(		(RTBL[0] != 'R')
            ||	(RTBL[1] != 'T')
            ||	(RTBL[2] != 'B')
            ||	(RTBL[3] != 'L')) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: No RTBL-Section found!\n");
        }

        RTBL_Length = SDL_SwapBE32( *((Uint32*) (RTBL + 4)));
        RTBL += 8;

        if(IcnFiledata + IcnFilesize < RTBL + RTBL_Length) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: RTBL-Section is bigger than ICN-File!\n");
        }

        NumFiles = SSET_Length / ((SIZE_X * SIZE_Y) / 2);

        if(RTBL_Length < NumFiles) {
            throw std::runtime_error("Icnfile::Icnfile(): Invalid ICN-File: RTBL-Section is too small!\n");
        }

        if(freesrc) SDL_RWclose(IcnRWop);
        if(freesrc) SDL_RWclose(MapRWop);
    } catch (std::exception &e) {
        delete [] MapFiledata;
        delete [] IcnFiledata;

        if(freesrc) SDL_RWclose(IcnRWop);
        if(freesrc) SDL_RWclose(MapRWop);
        throw;
    }
}

/// Destructor
/**
	Frees all memory.
*/
Icnfile::~Icnfile()
{
	delete [] IcnFiledata;
}

/// Returns one tile in the icn-File
/**
	This method returns a SDL_Surface containing the nth tile/picture in the icn-File.
	The returned SDL_Surface should be freed with SDL_FreeSurface() if no longer needed.
	\param	IndexOfFile	specifies which tile/picture to return (zero based)
	\return	nth tile/picture in this icn-File
*/
SDL_Surface* Icnfile::getPicture(Uint32 IndexOfFile) {
	SDL_Surface * pic;

	if(IndexOfFile >= NumFiles) {
		return NULL;
	}

	// check if palette is in range
	if(RTBL[IndexOfFile] >= RPAL_Length / 16) {
		return NULL;
	}

	unsigned char* palettestart = RPAL + (16 * RTBL[IndexOfFile]);

	unsigned char * filestart = SSET + (IndexOfFile * ((SIZE_X * SIZE_Y)/2));

	// create new picture surface
	if((pic = SDL_CreateRGBSurface(SDL_HWSURFACE,SIZE_X,SIZE_Y,8,0,0,0,0))== NULL) {
		return NULL;
	}

	palette.applyToSurface(pic);
	SDL_LockSurface(pic);

	//Now we can copy to surface
	unsigned char *dest = (unsigned char*) (pic->pixels);
	unsigned char pixel;
	for(int y = 0; y < SIZE_Y;y++) {
		for(int x = 0; x < SIZE_X; x+=2) {
			pixel = filestart[ (y*SIZE_X + x) / 2];
			pixel = pixel >> 4;
			dest[x] = palettestart[pixel];

			pixel = filestart[ (y*SIZE_X + x) / 2];
			pixel = pixel & 0x0F;
			dest[x+1] = palettestart[pixel];
		}
		dest += pic->pitch;
	}

	SDL_UnlockSurface(pic);

	return pic;
}

/// Returns an array of pictures in the icn-File
/**
	This method returns a SDL_Surface containing multiple tiles/pictures. Which tiles to include is specified by MapfileIndex. The
	MapfileIndex specifies the tileset. One tileset constists of multiple tiles of the icn-File.
	The last 3 parameters specify how to arrange the tiles:
	 - If all 3 parameters are 0 then a "random" layout is choosen, which should look good.
	 - If tilesX and tilesY is set to non-zero values then the result surface contains tilesX*tilesY tiles and this tilesN-times side by side.
	 - If all there parameters are non-zero then the result surface is exactly in this arrangement.

	tilesX*tilesY*tilesN must always the number of tiles in this tileset. Otherwise NULL is returned.<br><br>
	Example:
	\code
	Tileset = 10,11,12,13,14,15,16,17,18,19,20,21
	tilesX = 2; tilesY = 2; tilesN = 3

	returned picture:
	 10 11 14 15 18 19
	 12 13 16 17 20 21
	\endcode
	<br>
	The returned SDL_Surface should be freed with SDL_FreeSurface() if no longer needed.
	\param	MapfileIndex	specifies which tileset to use (zero based)
	\param	tilesX			how many tiles in x direction
	\param	tilesY			how many tiles in y direction
	\param	tilesN			how many tilesX*tilesY blocks in a row
	\return	the result surface with tilesX*tilesY*tilesN tiles
*/
SDL_Surface* Icnfile::getPictureArray(Uint32 MapfileIndex, int tilesX, int tilesY, int tilesN) {
	SDL_Surface * pic;

	if(MapfileIndex >= Tilesets.size()) {
		return NULL;
	}

	if((tilesX == 0) && (tilesY == 0) && (tilesN == 0)) {
		// guest what is best
		int tmp = Tilesets[MapfileIndex].NumTiles;
		if(tmp == 24) {
			// special case (radar station and light factory)
			tilesX = 2;
			tilesY = 2;
			tilesN = 6;
		} else if((tmp % 9) == 0) {
			tilesX = 3;
			tilesY = 3;
			tilesN = tmp / 9;
		} else if((tmp % 6) == 0) {
			tilesX = 3;
			tilesY = 2;
			tilesN = tmp / 6;
		} else if((tmp % 4) == 0) {
			tilesX = 2;
			tilesY = 2;
			tilesN = tmp / 4;
		} else if((tmp>=40) && ((tmp % 5) == 0)) {
			tilesX = tmp/5;
			tilesY = 5;
			tilesN = 1;
		} else {
			tilesX = 1;
			tilesY = 1;
			tilesN = tmp;
		}

	} else if( ((tilesX == 0) || (tilesY == 0)) && (tilesN == 0)) {
		// not possible
		return NULL;
	} else if((tilesX == 0) && (tilesY == 0) && (tilesN != 0)) {
		if(Tilesets[MapfileIndex].NumTiles % tilesN == 0) {
			// guest what is best
			int tmp = Tilesets[MapfileIndex].NumTiles / tilesN;
			if((tmp % 3) == 0) {
				tilesX = tmp/3;
				tilesY = 3;
			} else if((tmp % 2) == 0) {
				tilesX = tmp/2;
				tilesY = 2;
			} else {
				tilesX = tmp;
				tilesY = 1;
			}
		} else {
			// not possible
			return NULL;
		}
	} else {
		if((unsigned int)tilesX*tilesY*tilesN != Tilesets[MapfileIndex].NumTiles) {
			return NULL;
		}
	}

	// create new picture surface
	if((pic = SDL_CreateRGBSurface(SDL_HWSURFACE,SIZE_X*tilesX*tilesN,SIZE_Y*tilesY,8,0,0,0,0))== NULL) {
		return NULL;
	}

	palette.applyToSurface(pic);
	SDL_LockSurface(pic);

	int Tileidx=0;
	for(int n = 0; n < tilesN; n++) {
		for(int y = 0; y < tilesY; y++) {
			for(int x = 0; x < tilesX; x++) {
				int IndexOfFile = Tilesets[MapfileIndex].TileIndices[Tileidx];

				// check if palette is in range
				if(RTBL[IndexOfFile] >= RPAL_Length / 16) {
					SDL_UnlockSurface(pic);
					SDL_FreeSurface(pic);
					return NULL;
				}

				unsigned char* palettestart = RPAL + (16 * RTBL[IndexOfFile]);
				unsigned char * filestart = SSET + (IndexOfFile * ((SIZE_X * SIZE_Y)/2));

				//Now we can copy to surface
				unsigned char *dest = (unsigned char*) (pic->pixels) + (pic->pitch)*y*SIZE_Y + (x+n*tilesX) * SIZE_X;
				unsigned char pixel;
				for(int y = 0; y < SIZE_Y;y++) {
					for(int x = 0; x < SIZE_X; x+=2) {
						pixel = filestart[ (y*SIZE_X + x) / 2];
						pixel = pixel >> 4;
						dest[x] = palettestart[pixel];

						pixel = filestart[ (y*SIZE_X + x) / 2];
						pixel = pixel & 0x0F;
						dest[x+1] = palettestart[pixel];
					}
					dest += pic->pitch;
				}

				Tileidx++;
			}
		}
	}

	SDL_UnlockSurface(pic);

	return pic;
}

/// Returns a row of pictures in the icn-File
/**
	This method returns a SDL_Surface containing multiple tiles/pictures. The returned surface contains all
	tiles from StartIndex to EndIndex.
	The returned SDL_Surface should be freed with SDL_FreeSurface() if no longer needed.
	\param	StartIndex		The first tile to use
	\param	EndIndex		The last tile to use
	\return	the result surface with (EndIndex-StartIndex+1) tiles. NULL on errors.
*/
SDL_Surface* Icnfile::getPictureRow(Uint32 StartIndex, Uint32 EndIndex) {
	SDL_Surface * pic;

	if((StartIndex >= NumFiles)||(EndIndex >= NumFiles)||(StartIndex > EndIndex)) {
		return NULL;
	}

	Uint32 NumTiles = EndIndex - StartIndex + 1;
	// create new picture surface
	if((pic = SDL_CreateRGBSurface(SDL_HWSURFACE,SIZE_X*NumTiles,SIZE_Y,8,0,0,0,0))== NULL) {
		return NULL;
	}

	palette.applyToSurface(pic);
	SDL_LockSurface(pic);

	for(Uint32 i = 0; i < NumTiles; i++) {
		int IndexOfFile = i+StartIndex;

		// check if palette is in range
		if(RTBL[IndexOfFile] >= RPAL_Length / 16) {
			SDL_UnlockSurface(pic);
			SDL_FreeSurface(pic);
			return NULL;
		}

		unsigned char* palettestart = RPAL + (16 * RTBL[IndexOfFile]);
		unsigned char * filestart = SSET + (IndexOfFile * ((SIZE_X * SIZE_Y)/2));

		//Now we can copy to surface
		unsigned char *dest = (unsigned char*) (pic->pixels) + i*SIZE_X;
		unsigned char pixel;
		for(int y = 0; y < SIZE_Y;y++) {
			for(int x = 0; x < SIZE_X; x+=2) {
				pixel = filestart[ (y*SIZE_X + x) / 2];
				pixel = pixel >> 4;
				dest[x] = palettestart[pixel];

				pixel = filestart[ (y*SIZE_X + x) / 2];
				pixel = pixel & 0x0F;
				dest[x+1] = palettestart[pixel];
			}
			dest += pic->pitch;
		}
	}

	SDL_UnlockSurface(pic);
	return pic;
}

/// Returns a row of pictures in the icn-File
/**
	This method returns a SDL_Surface containing multiple tiles/pictures. The returned surface contains all
	tiles specified be the parameters
	The returned SDL_Surface should be freed with SDL_FreeSurface() if no longer needed.
	\param	numTiles    the number of tiles that should be extrated.
	\return	the result surface with all specified tiles in a row. NULL on errors.
*/
SDL_Surface* Icnfile::getPictureRow2(unsigned int numTiles, ...) {
	SDL_Surface * pic;

	// create new picture surface
	if((pic = SDL_CreateRGBSurface(SDL_HWSURFACE,SIZE_X*numTiles,SIZE_Y,8,0,0,0,0))== NULL) {
		return NULL;
	}

	palette.applyToSurface(pic);
	SDL_LockSurface(pic);

    va_list arg_ptr;
	va_start(arg_ptr, numTiles);

	for(unsigned int i = 0; i < numTiles; i++) {
		unsigned int IndexOfFile = va_arg( arg_ptr, unsigned int);

        if((IndexOfFile >= NumFiles) || (IndexOfFile >= NumFiles)) {
            SDL_UnlockSurface(pic);
            SDL_FreeSurface(pic);
            va_end(arg_ptr);
            return NULL;
        }

		// check if palette is in range
		if(RTBL[IndexOfFile] >= RPAL_Length / 16) {
			SDL_UnlockSurface(pic);
			SDL_FreeSurface(pic);
			va_end(arg_ptr);
			return NULL;
		}

		unsigned char* palettestart = RPAL + (16 * RTBL[IndexOfFile]);
		unsigned char * filestart = SSET + (IndexOfFile * ((SIZE_X * SIZE_Y)/2));

		//Now we can copy to surface
		unsigned char *dest = (unsigned char*) (pic->pixels) + i*SIZE_X;
		unsigned char pixel;
		for(int y = 0; y < SIZE_Y;y++) {
			for(int x = 0; x < SIZE_X; x+=2) {
				pixel = filestart[ (y*SIZE_X + x) / 2];
				pixel = pixel >> 4;
				dest[x] = palettestart[pixel];

				pixel = filestart[ (y*SIZE_X + x) / 2];
				pixel = pixel & 0x0F;
				dest[x+1] = palettestart[pixel];
			}
			dest += pic->pitch;
		}
	}

    va_end(arg_ptr);

	SDL_UnlockSurface(pic);
	return pic;
}
