/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <GUI/TextView.h>

#include <algorithm>

TextView::TextView() : Widget() {
    fontID = FONT_STD12;
	textcolor = -1;
	textshadowcolor = -1;
	backgroundcolor = 0;
	alignment = (Alignment_Enum) (Alignment_Left | Alignment_Top);
	pForeground = NULL;
	pBackground = NULL;
	bAutohideScrollbar = true;
	EnableResizing(true,true);

	Resize(GetMinimumSize().x,GetMinimumSize().y);
}

TextView::~TextView() {
	if(pBackground != NULL) {
		SDL_FreeSurface(pBackground);
	}

	if(pForeground != NULL) {
		SDL_FreeSurface(pForeground);
	}
}

void TextView::HandleMouseMovement(Sint32 x, Sint32 y, bool insideOverlay) {
	scrollbar.HandleMouseMovement(x - GetSize().x + scrollbar.GetSize().x,y,insideOverlay);
}

bool TextView::HandleMouseLeft(Sint32 x, Sint32 y, bool pressed) {
	return scrollbar.HandleMouseLeft(x - GetSize().x + scrollbar.GetSize().x,y,pressed);
}

bool TextView::HandleMouseWheel(Sint32 x, Sint32 y, bool up)  {
	// forward mouse wheel event to scrollbar
	return scrollbar.HandleMouseWheel(0,0,up);
}

bool TextView::HandleKeyPress(SDL_KeyboardEvent& key) {
	Widget::HandleKeyPress(key);

	scrollbar.HandleKeyPress(key);
	return true;
}

void TextView::Draw(SDL_Surface* screen, Point Position) {
	if(IsVisible() == false) {
		return;
	}

	if(pBackground != NULL) {
		SDL_Rect dest;
		dest.x = Position.x;
		dest.y = Position.y;
		dest.w = pBackground->w;
		dest.h = pBackground->h;

		SDL_BlitSurface(pBackground,NULL,screen,&dest);
	}

	int lineHeight = GUIStyle::GetInstance().GetTextHeight(fontID) + 2;

	SDL_Rect src;
	src.x = 0;
	src.y = scrollbar.GetCurrentValue() * lineHeight;
	src.w = pForeground->w;
	src.h = GetSize().y - 2;

	SDL_Rect dest;
	dest.x = Position.x + 2;
	dest.y = Position.y + 1;
	dest.w = pForeground->w;
	dest.h = pForeground->h;
	SDL_BlitSurface(pForeground,&src,screen,&dest);

	Point ScrollBarPos = Position;
	ScrollBarPos.x += GetSize().x - scrollbar.GetSize().x;

    if(!bAutohideScrollbar || (scrollbar.GetRangeMin() != scrollbar.GetRangeMax())) {
        scrollbar.Draw(screen,ScrollBarPos);
    }
}

void TextView::Resize(Uint32 width, Uint32 height) {
	if(pForeground != NULL) {
		SDL_FreeSurface(pForeground);
		pForeground = NULL;
	}

	if(pBackground != NULL) {
		SDL_FreeSurface(pBackground);
	}

	pBackground = GUIStyle::GetInstance().CreateWidgetBackground(width, height);

	scrollbar.Resize(scrollbar.GetMinimumSize().x,height);

	//split text into single lines at every '\n'
	size_t startpos = 0;
	size_t nextpos;
	std::list<std::string> HardLines;
	do {
		nextpos = Text.find("\n",startpos);
		if(nextpos == std::string::npos) {
			HardLines.push_back(Text.substr(startpos,Text.length()-startpos));
		} else {
			HardLines.push_back(Text.substr(startpos,nextpos-startpos));
			startpos = nextpos+1;
		}
	} while(nextpos != std::string::npos);

	std::list<std::string> TextLines;

	std::list<std::string>::const_iterator iter;
	for(iter = HardLines.begin();iter != HardLines.end();++iter) {
		std::string tmpLine = *iter;

		if(tmpLine == "") {
            TextLines.push_back(" ");
            continue;
		}

		bool EndOfLine = false;
		size_t warppos = 0;
		size_t oldwarppos = 0;
		size_t lastwarp = 0;

		while(EndOfLine == false) {
			while(true) {
				warppos = tmpLine.find(" ", oldwarppos);
				std::string tmp;
				if(warppos == std::string::npos) {
					tmp = tmpLine.substr(lastwarp,tmpLine.length()-lastwarp);
					warppos = tmpLine.length();
					EndOfLine = true;
				} else {
					tmp = tmpLine.substr(lastwarp,warppos-lastwarp);
				}

				if( (int) GUIStyle::GetInstance().GetMinimumLabelSize(tmp, fontID).x - 4 > (int) width - scrollbar.GetSize().x - 4) {
					// this line would be too big => in oldwarppos is the last correct warp pos
					EndOfLine = false;
					break;
				} else {
				    if(EndOfLine == true) {
                        oldwarppos = warppos;
                        break;
				    } else {
                        oldwarppos = warppos + 1;
				    }
				}
			}

			if(oldwarppos == lastwarp) {
				// the width of this label is too small for the next word
				// split the word
				
				warppos = lastwarp;
				while(true) {
					std::string tmp = tmpLine.substr(lastwarp,warppos-lastwarp);
					if( (int) GUIStyle::GetInstance().GetMinimumLabelSize(tmp, fontID).x - 4 > (int) width - scrollbar.GetSize().x - 4) {
						// this line would be too big => in oldwarppos is the last correct warp pos
						break;
					} else {
						oldwarppos = warppos;
					}

					warppos++;
				}

				if(warppos != lastwarp) {
					TextLines.push_back(tmpLine.substr(lastwarp,oldwarppos-lastwarp));
					lastwarp = oldwarppos;
				} else {
					// the width of this label is too small for the next character
					// create a dummy entry
					TextLines.push_back(" ");
					lastwarp++;
					oldwarppos++;
				}
			} else {
				std::string tmpStr = tmpLine.substr(lastwarp,oldwarppos-lastwarp);
				TextLines.push_back(tmpStr);
				lastwarp = oldwarppos;
			}
		}
	}

	int lineHeight = GUIStyle::GetInstance().GetTextHeight(fontID) + 2;
	int labelHeight = lineHeight * TextLines.size() + 2;
	pForeground = GUIStyle::GetInstance().CreateLabelSurface(width-4,labelHeight,TextLines,fontID,alignment,textcolor,textshadowcolor,backgroundcolor);

	int numVisibleLines = height/lineHeight;
	scrollbar.SetRange(0,std::max(0, ((int) TextLines.size()) - numVisibleLines));
	scrollbar.SetBigStepSize(std::max(1, numVisibleLines-1));

	Widget::Resize(width,height);
}


