/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <Menu/CustomGamePlayers.h>

#include <FileClasses/GFXManager.h>
#include <FileClasses/TextManager.h>
#include <FileClasses/INIFile.h>

#include <Network/NetworkManager.h>

#include <GUI/Spacer.h>
#include <GUI/GUIStyle.h>
#include <GUI/MsgBox.h>

#include <misc/fnkdat.h>
#include <misc/FileSystem.h>
#include <misc/draw_util.h>
#include <misc/string_util.h>
#include <misc/IMemoryStream.h>

#include <INIMap/INIMapPreviewCreator.h>

#include <sand.h>
#include <globals.h>


#define PLAYER_OPEN         0
#define PLAYER_CLOSED       1
#define PLAYER_HUMAN        2
#define PLAYER_AI_EASY      -1
#define PLAYER_AI_MEDIUM    -2
#define PLAYER_AI_DIFFICULT -3


CustomGamePlayers::CustomGamePlayers(const GameInitSettings& newGameInitSettings, bool server, bool LANServer)
 : MenuBase(), gameInitSettings(newGameInitSettings), bServer(server), bLANServer(LANServer), startGameTime(0) {

	// set up window
	SDL_Surface *surf;
	surf = pGFXManager->getUIGraphic(UI_MenuBackground);

	SetBackground(surf,false);
	Resize(surf->w,surf->h);

	SetWindowWidget(&WindowWidget);

	WindowWidget.AddWidget(&MainVBox, Point(24,23),	Point(screen->w - 48, screen->h - 32));

    CaptionLabel.SetText(extractFilenameFromPath(gameInitSettings.getFilename()));
    CaptionLabel.SetAlignment(Alignment_HCenter);
    MainVBox.AddWidget(&CaptionLabel, 24);
    MainVBox.AddWidget(VSpacer::Create(24));

    MainVBox.AddWidget(Spacer::Create(), 0.04);

    MainVBox.AddWidget(&MainHBox, 0.6);

    MainHBox.AddWidget(Spacer::Create(), 0.05);
    LeftVBox.AddWidget(Spacer::Create(), 0.1);
    LeftVBox.AddWidget(&PlayerListHBox);
    PlayerListHBox.AddWidget(Spacer::Create(), 0.4);
    PlayerListHBox.AddWidget(&PlayerListVBox, 0.2);
    PlayerListHBox.AddWidget(Spacer::Create(), 0.4);
    MainHBox.AddWidget(&LeftVBox, 0.8);

    MainHBox.AddWidget(Spacer::Create(), 0.05);

    MainHBox.AddWidget(HSpacer::Create(8));

    MainHBox.AddWidget(&RightVBox, 180);
    MainHBox.AddWidget(Spacer::Create(), 0.05);
    Minimap.SetSurface( GUIStyle::GetInstance().CreateButtonSurface(130,130,pTextManager->getLocalized("Choose map"), true, false), true);
    RightVBox.AddWidget(&Minimap);

    if(gameInitSettings.getGameType() == GAMETYPE_CUSTOM_MULTIPLAYER) {
        SDL_RWops* RWops = SDL_RWFromConstMem(gameInitSettings.getFiledata().c_str(), gameInitSettings.getFiledata().size());

        std::shared_ptr<INIFile> map(new INIFile(RWops));
        extractMapInfo(map);

        SDL_RWclose(RWops);
    } else if(gameInitSettings.getGameType() == GAMETYPE_LOAD_MULTIPLAYER) {
        IMemoryStream memStream(gameInitSettings.getFiledata().c_str(), gameInitSettings.getFiledata().size());

        Uint32 magicNum = memStream.readUint32();
        if(magicNum != SAVEMAGIC) {
            fprintf(stderr,"CustomGamePlayers: No valid savegame! Expected magic number %.8X, but got %.8X!\n", SAVEMAGIC, magicNum);
        }

        Uint32 savegameVersion = memStream.readUint32();
        if (savegameVersion != SAVEGAMEVERSION) {
            fprintf(stderr,"CustomGamePlayers: No valid savegame! Expected savegame version %d, but got %d!\n", SAVEGAMEVERSION, savegameVersion);
        }

        memStream.readString();     // dune legacy version

        // read gameInitSettings
        GameInitSettings tmpGameInitSettings(memStream);

        Uint32 numHouseInfo = memStream.readUint32();
        for(Uint32 i=0;i<numHouseInfo;i++) {
            houseInfoListSetup.push_back(GameInitSettings::HouseInfo(memStream));
        }

        SDL_RWops* RWops = SDL_RWFromConstMem(tmpGameInitSettings.getFiledata().c_str(), tmpGameInitSettings.getFiledata().size());

        std::shared_ptr<INIFile> map(new INIFile(RWops));
        extractMapInfo(map);

        SDL_RWclose(RWops);

        // adjust multiple players per house as this was not known before actually loading the saved game
        gameInitSettings.setMultiplePlayersPerHouse(tmpGameInitSettings.isMultiplePlayersPerHouse());

        // adjust numHouses to the actually used houses (which might be smaller than the houses on the map)
        numHouses = houseInfoListSetup.size();
    } else {
        std::shared_ptr<INIFile> map(new INIFile(gameInitSettings.getFilename()));
        extractMapInfo(map);
    }

    RightVBox.AddWidget(VSpacer::Create(10));
    RightVBox.AddWidget(&MapPropertiesHBox, 0.01);
    MapPropertiesHBox.AddWidget(&MapPropertyNamesVBox, 75);
    MapPropertiesHBox.AddWidget(&MapPropertyValuesVBox, 105);
    MapPropertyNamesVBox.AddWidget(Label::Create(pTextManager->getLocalized("Size") + ":"));
    MapPropertyValuesVBox.AddWidget(&MapProperty_Size);
    MapPropertyNamesVBox.AddWidget(Label::Create(pTextManager->getLocalized("Players") + ":"));
    MapPropertyValuesVBox.AddWidget(&MapProperty_Players);
    MapPropertyNamesVBox.AddWidget(Label::Create(pTextManager->getLocalized("Author") + ":"));
    MapPropertyValuesVBox.AddWidget(&MapProperty_Authors);
    MapPropertyNamesVBox.AddWidget(Label::Create(pTextManager->getLocalized("License") + ":"));
    MapPropertyValuesVBox.AddWidget(&MapProperty_License);
    RightVBox.AddWidget(Spacer::Create());

    MainVBox.AddWidget(Spacer::Create(), 0.04);

	MainVBox.AddWidget(&ButtonHBox, 0.32);

	ButtonHBox.AddWidget(HSpacer::Create(70));

	Button_BackVBox.AddWidget(Spacer::Create());
	Button_Back.SetText(pTextManager->getLocalized("Back"));
	Button_Back.SetOnClick(std::bind(&CustomGamePlayers::OnCancel, this));
	Button_BackVBox.AddWidget(&Button_Back, 24);
	Button_BackVBox.AddWidget(VSpacer::Create(14));
	ButtonHBox.AddWidget(&Button_BackVBox, 0.1);

    ButtonHBox.AddWidget(Spacer::Create(), 0.0625);

    ChatTextView.SetTextFont(FONT_STD10);
	ChatVBox.AddWidget(&ChatTextView, 0.77);
	if(screen->h <= 600) {
        ChatTextBox.SetTextFont(FONT_STD10);
	}
	ChatTextBox.SetOnReturn(std::bind(&CustomGamePlayers::OnSendChatMessage, this));
	ChatVBox.AddWidget(&ChatTextBox, 0.2);
	ChatVBox.AddWidget(Spacer::Create(), 0.03);
	ButtonHBox.AddWidget(&ChatVBox, 0.675);

	if(gameInitSettings.getGameType() != GAMETYPE_CUSTOM_MULTIPLAYER && gameInitSettings.getGameType() != GAMETYPE_LOAD_MULTIPLAYER) {
        ChatVBox.SetVisible(false);
        ChatVBox.SetEnabled(false);
	}

	bool bLoadMultiplayer = (gameInitSettings.getGameType() == GAMETYPE_LOAD_MULTIPLAYER);

	ButtonHBox.AddWidget(Spacer::Create(), 0.0625);

	Button_NextVBox.AddWidget(Spacer::Create());
    Button_Next.SetText(pTextManager->getLocalized("Next"));
	Button_Next.SetOnClick(std::bind(&CustomGamePlayers::OnNext, this));
	Button_NextVBox.AddWidget(&Button_Next, 24);
	Button_NextVBox.AddWidget(VSpacer::Create(14));
	if(bServer == false) {
        Button_Next.SetEnabled(false);
        Button_Next.SetVisible(false);
	}
	ButtonHBox.AddWidget(&Button_NextVBox, 0.1);

	ButtonHBox.AddWidget(HSpacer::Create(90));

    std::list<HOUSETYPE>  tmpBoundHousesOnMap = boundHousesOnMap;

    bool thisPlayerPlaced = false;

    for(int i=0;i<NUM_HOUSES;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        // set up header row with Label "House", DropDown for house selection and DropDown for team selection
        curHouseInfo.HouseLabel.SetText(pTextManager->getLocalized("House"));
        curHouseInfo.HouseHBox.AddWidget(&curHouseInfo.HouseLabel, 60);

        if(bLoadMultiplayer) {
            if(i < (int) houseInfoListSetup.size()) {
                GameInitSettings::HouseInfo gisHouseInfo = houseInfoListSetup.at(i);

                if(gisHouseInfo.HouseID == HOUSE_INVALID) {
                    addToHouseDropDown(curHouseInfo.HouseDropDown, HOUSE_INVALID, true);
                } else {
                    addToHouseDropDown(curHouseInfo.HouseDropDown, gisHouseInfo.HouseID, true);
                }
            }

            curHouseInfo.HouseDropDown.SetEnabled(false);
        } else if(tmpBoundHousesOnMap.empty() == false) {
            HOUSETYPE housetype = tmpBoundHousesOnMap.front();
            tmpBoundHousesOnMap.pop_front();

            addToHouseDropDown(curHouseInfo.HouseDropDown, housetype, true);
            curHouseInfo.HouseDropDown.SetEnabled(bServer);
        } else {
            addToHouseDropDown(curHouseInfo.HouseDropDown, HOUSE_INVALID, true);
            curHouseInfo.HouseDropDown.SetEnabled(bServer);
        }
        curHouseInfo.HouseDropDown.SetOnSelectionChange(std::bind(&CustomGamePlayers::OnChangeHousesDropDownBoxes, this, std::placeholders::_1, i));
        curHouseInfo.HouseHBox.AddWidget(&curHouseInfo.HouseDropDown, 95);

        if(bLoadMultiplayer) {
            if(i < (int) houseInfoListSetup.size()) {
                GameInitSettings::HouseInfo gisHouseInfo = houseInfoListSetup.at(i);

                curHouseInfo.TeamDropDown.AddEntry(pTextManager->getLocalized("Team") + " " + stringify(gisHouseInfo.team), gisHouseInfo.team);
                curHouseInfo.TeamDropDown.SetSelectedItem(0);
            }
            curHouseInfo.TeamDropDown.SetEnabled(false);
            curHouseInfo.TeamDropDown.SetOnClickEnabled(false);
        } else {
            for(int team = 0 ; team<numHouses ; team++) {
                curHouseInfo.TeamDropDown.AddEntry(pTextManager->getLocalized("Team") + " " + stringify(team+1), team+1);
            }
            curHouseInfo.TeamDropDown.SetSelectedItem(i);
            curHouseInfo.TeamDropDown.SetEnabled(bServer);
        }
        curHouseInfo.TeamDropDown.SetOnSelectionChange(std::bind(&CustomGamePlayers::OnChangeTeamDropDownBoxes, this, std::placeholders::_1, i));
        curHouseInfo.HouseHBox.AddWidget(HSpacer::Create(10));
        curHouseInfo.HouseHBox.AddWidget(&curHouseInfo.TeamDropDown, 85);

        curHouseInfo.HouseInfoVBox.AddWidget(&curHouseInfo.HouseHBox);

        // add 1. player
        curHouseInfo.Player1ArrowLabel.SetSurface(pGFXManager->getUIGraphic(UI_CustomGamePlayersArrowNeutral), false);
        curHouseInfo.PlayerHBox.AddWidget(&curHouseInfo.Player1ArrowLabel);
        curHouseInfo.Player1Label.SetText(pTextManager->getLocalized("Player") + (gameInitSettings.isMultiplePlayersPerHouse() ? " 1" : ""));
        curHouseInfo.Player1Label.SetTextFont(FONT_STD10);
        curHouseInfo.PlayerHBox.AddWidget(&curHouseInfo.Player1Label, 68);

        if(bLoadMultiplayer) {
            if(i < (int) houseInfoListSetup.size()) {
                GameInitSettings::HouseInfo gisHouseInfo = houseInfoListSetup.at(i);

                if(gisHouseInfo.HouseID == HOUSE_UNUSED) {
                    curHouseInfo.Player1DropDown.AddEntry(pTextManager->getLocalized("closed"), PLAYER_CLOSED);
                } else if(gisHouseInfo.playerInfoList.empty() == false) {
                    GameInitSettings::PlayerInfo playerInfo = gisHouseInfo.playerInfoList.front();
                    if(playerInfo.PlayerClass == "HumanPlayer") {
                        if(thisPlayerPlaced == false) {
                            curHouseInfo.Player1DropDown.AddEntry(settings.General.PlayerName, PLAYER_HUMAN);
                            curHouseInfo.Player1DropDown.SetSelectedItem(0);
                            thisPlayerPlaced = true;
                        } else {
                            curHouseInfo.Player1DropDown.AddEntry(pTextManager->getLocalized("open"), PLAYER_OPEN);
                            curHouseInfo.Player1DropDown.SetSelectedItem(0);
                        }
                    } else if(playerInfo.PlayerClass == "AIPlayer") {
                        if(playerInfo.PlayerClassParameter == "0") {
                            curHouseInfo.Player1DropDown.AddEntry("AI (easy)", PLAYER_AI_EASY);
                            curHouseInfo.Player1DropDown.SetSelectedItem(0);
                            curHouseInfo.Player1DropDown.SetEnabled(false);
                            curHouseInfo.Player1DropDown.SetOnClickEnabled(false);
                        } else if(playerInfo.PlayerClassParameter == "1") {
                            curHouseInfo.Player1DropDown.AddEntry("AI (medium)", PLAYER_AI_MEDIUM);
                            curHouseInfo.Player1DropDown.SetSelectedItem(0);
                            curHouseInfo.Player1DropDown.SetEnabled(false);
                            curHouseInfo.Player1DropDown.SetOnClickEnabled(false);
                        } else if(playerInfo.PlayerClassParameter == "2") {
                            curHouseInfo.Player1DropDown.AddEntry("AI (difficult)", PLAYER_AI_DIFFICULT);
                            curHouseInfo.Player1DropDown.SetSelectedItem(0);
                            curHouseInfo.Player1DropDown.SetEnabled(false);
                            curHouseInfo.Player1DropDown.SetOnClickEnabled(false);
                        }
                    }
                }

            }

        } else if(bServer && (thisPlayerPlaced == false)) {
            curHouseInfo.Player1DropDown.AddEntry(settings.General.PlayerName, PLAYER_HUMAN);
            curHouseInfo.Player1DropDown.SetSelectedItem(0);
            curHouseInfo.Player1DropDown.SetEnabled(bServer);
            thisPlayerPlaced = true;
        } else {
            curHouseInfo.Player1DropDown.AddEntry(pTextManager->getLocalized("open"), PLAYER_OPEN);
            curHouseInfo.Player1DropDown.SetSelectedItem(0);
            curHouseInfo.Player1DropDown.AddEntry(pTextManager->getLocalized("closed"), PLAYER_CLOSED);
            curHouseInfo.Player1DropDown.AddEntry("AI (easy)", PLAYER_AI_EASY);
            curHouseInfo.Player1DropDown.AddEntry("AI (medium)", PLAYER_AI_MEDIUM);
            curHouseInfo.Player1DropDown.AddEntry("AI (difficult)", PLAYER_AI_DIFFICULT);
            curHouseInfo.Player1DropDown.SetEnabled(bServer);
        }
        curHouseInfo.Player1DropDown.SetOnSelectionChange(std::bind(&CustomGamePlayers::OnChangePlayerDropDownBoxes, this, std::placeholders::_1, 2*i));
        curHouseInfo.Player1DropDown.SetOnClick(std::bind(&CustomGamePlayers::OnClickPlayerDropDownBox, this, 2*i));
        curHouseInfo.PlayerHBox.AddWidget(&curHouseInfo.Player1DropDown, 100);

        curHouseInfo.PlayerHBox.AddWidget(HSpacer::Create(10));

        // add 2. player
        curHouseInfo.Player2Label.SetText(pTextManager->getLocalized("Player") + " 2");
        curHouseInfo.Player2Label.SetTextFont(FONT_STD10);
        curHouseInfo.PlayerHBox.AddWidget(&curHouseInfo.Player2Label, 68);

        if(bLoadMultiplayer) {
            if(i < (int) houseInfoListSetup.size()) {
                GameInitSettings::HouseInfo gisHouseInfo = houseInfoListSetup.at(i);

                if(gisHouseInfo.HouseID == HOUSE_UNUSED) {
                    curHouseInfo.Player2DropDown.AddEntry(pTextManager->getLocalized("closed"), PLAYER_CLOSED);
                } else if(gisHouseInfo.playerInfoList.size() >= 2) {
                    GameInitSettings::PlayerInfo playerInfo = *(++gisHouseInfo.playerInfoList.begin());
                    if(playerInfo.PlayerClass == "HumanPlayer") {
                        if(thisPlayerPlaced == false) {
                            curHouseInfo.Player2DropDown.AddEntry(settings.General.PlayerName, PLAYER_HUMAN);
                            curHouseInfo.Player2DropDown.SetSelectedItem(0);
                            thisPlayerPlaced = true;
                        } else {
                            curHouseInfo.Player2DropDown.AddEntry(pTextManager->getLocalized("open"), PLAYER_OPEN);
                            curHouseInfo.Player2DropDown.SetSelectedItem(0);
                        }
                    } else if(playerInfo.PlayerClass == "AIPlayer") {
                        if(playerInfo.PlayerClassParameter == "0") {
                            curHouseInfo.Player2DropDown.AddEntry("AI (easy)", PLAYER_AI_EASY);
                            curHouseInfo.Player2DropDown.SetSelectedItem(0);
                            curHouseInfo.Player2DropDown.SetEnabled(false);
                            curHouseInfo.Player2DropDown.SetOnClickEnabled(false);
                        } else if(playerInfo.PlayerClassParameter == "1") {
                            curHouseInfo.Player2DropDown.AddEntry("AI (medium)", PLAYER_AI_MEDIUM);
                            curHouseInfo.Player2DropDown.SetSelectedItem(0);
                            curHouseInfo.Player2DropDown.SetEnabled(false);
                            curHouseInfo.Player2DropDown.SetOnClickEnabled(false);
                        } else if(playerInfo.PlayerClassParameter == "2") {
                            curHouseInfo.Player2DropDown.AddEntry("AI (difficult)", PLAYER_AI_DIFFICULT);
                            curHouseInfo.Player2DropDown.SetSelectedItem(0);
                            curHouseInfo.Player2DropDown.SetEnabled(false);
                            curHouseInfo.Player2DropDown.SetOnClickEnabled(false);
                        }
                    }
                }
            }
        } else if(bServer && (thisPlayerPlaced == false)) {
            curHouseInfo.Player2DropDown.AddEntry(settings.General.PlayerName, PLAYER_HUMAN);
            curHouseInfo.Player2DropDown.SetSelectedItem(0);
            curHouseInfo.Player2DropDown.SetEnabled(bServer);
            thisPlayerPlaced = true;
        } else {
            curHouseInfo.Player2DropDown.AddEntry(pTextManager->getLocalized("open"), PLAYER_OPEN);
            curHouseInfo.Player2DropDown.SetSelectedItem(0);
            curHouseInfo.Player2DropDown.AddEntry(pTextManager->getLocalized("closed"), PLAYER_CLOSED);
            curHouseInfo.Player2DropDown.AddEntry("AI (easy)", PLAYER_AI_EASY);
            curHouseInfo.Player2DropDown.AddEntry("AI (medium)", PLAYER_AI_MEDIUM);
            curHouseInfo.Player2DropDown.AddEntry("AI (difficult)", PLAYER_AI_DIFFICULT);
            curHouseInfo.Player2DropDown.SetEnabled(bServer);
        }
        curHouseInfo.Player2DropDown.SetOnSelectionChange(std::bind(&CustomGamePlayers::OnChangePlayerDropDownBoxes, this, std::placeholders::_1, 2*i + 1));
        curHouseInfo.Player2DropDown.SetOnClick(std::bind(&CustomGamePlayers::OnClickPlayerDropDownBox, this, 2*i + 1));
        curHouseInfo.PlayerHBox.AddWidget(&curHouseInfo.Player2DropDown, 100);

        curHouseInfo.HouseInfoVBox.AddWidget(&curHouseInfo.PlayerHBox);

        PlayerListVBox.AddWidget(&curHouseInfo.HouseInfoVBox, 0.01);

        PlayerListVBox.AddWidget(VSpacer::Create(4), 0.0);
        PlayerListVBox.AddWidget(Spacer::Create(), 0.07);

        if(i >= numHouses) {
            curHouseInfo.HouseInfoVBox.SetEnabled(false);
            curHouseInfo.HouseInfoVBox.SetVisible(false);
        }
    }

    OnChangeHousesDropDownBoxes(false);

    checkPlayerBoxes();

	LeftVBox.AddWidget(Spacer::Create(), 0.3);

    if(pNetworkManager != NULL) {
        if(bServer) {
            pNetworkManager->startServer(bLANServer, gameInitSettings.getServername(), settings.General.PlayerName, &gameInitSettings, 1, gameInitSettings.isMultiplePlayersPerHouse() ? numHouses*2 : numHouses);
        }

        pNetworkManager->SetOnPeerDisconnected(std::bind(&CustomGamePlayers::OnPeerDisconnected, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3));
        pNetworkManager->SetOnReceiveChangeEventList(std::bind(&CustomGamePlayers::OnReceiveChangeEventList, this, std::placeholders::_1));
        pNetworkManager->SetOnReceiveChatMessage(std::bind(&CustomGamePlayers::OnReceiveChatMessage, this, std::placeholders::_1, std::placeholders::_2));

        if(bServer) {
            pNetworkManager->SetGetChangeEventListForNewPlayerCallback(std::bind(&CustomGamePlayers::GetChangeEventListForNewPlayer, this, std::placeholders::_1));
        } else {
            pNetworkManager->SetOnStartGame(std::bind(&CustomGamePlayers::OnStartGame, this, std::placeholders::_1));
        }
    }
}

CustomGamePlayers::~CustomGamePlayers()
{
    if(pNetworkManager != NULL) {
        pNetworkManager->disconnect();

        pNetworkManager->SetOnPeerDisconnected(std::function<void (std::string, bool, int)>());
        pNetworkManager->SetGetChangeEventListForNewPlayerCallback(std::function<ChangeEventList (std::string)>());
        pNetworkManager->SetOnReceiveChangeEventList(std::function<void (ChangeEventList)>());
        pNetworkManager->SetOnReceiveChatMessage(std::function<void (std::string, std::string)>());
        pNetworkManager->SetOnStartGame(std::function<void (unsigned int)>());

        if(bServer) {
            pNetworkManager->stopServer();
        }
    }
}

void CustomGamePlayers::update() {
    if(startGameTime > 0) {
        if(SDL_GetTicks() >= startGameTime) {
            startGameTime = 0;

            pNetworkManager->SetOnPeerDisconnected(std::function<void (std::string, bool, int)>());
            pNetworkManager->SetGetChangeEventListForNewPlayerCallback(std::function<ChangeEventList (std::string)>());
            pNetworkManager->SetOnReceiveChangeEventList(std::function<void (ChangeEventList)>());
            pNetworkManager->SetOnReceiveChatMessage(std::function<void (std::string, std::string)>());
            pNetworkManager->SetOnStartGame(std::function<void (unsigned int)>());

            if(bServer) {
                pNetworkManager->stopServer();
            }

            addAllPlayersToGameInitSettings();
            startMultiPlayerGame(gameInitSettings);

            quit(MENU_QUIT_GAME_FINISHED);
        } else {
            static Uint32 lastSecondLeft = 0;
            Uint32 secondsLeft = ((startGameTime - SDL_GetTicks())/1000) + 1;
            if(lastSecondLeft != secondsLeft) {
                lastSecondLeft = secondsLeft;
                addInfoMessage("Starting game in " + stringify(secondsLeft) + "...");
            }
        }
    }
}

void CustomGamePlayers::OnReceiveChangeEventList(ChangeEventList changeEventList)
{
    std::list<ChangeEventList::ChangeEvent>::const_iterator iter;
    for(iter = changeEventList.changeEventList.begin(); iter != changeEventList.changeEventList.end(); ++iter) {

        switch(iter->eventType) {
            case ChangeEventList::ChangeEvent::EventType_ChangeHouse: {
                HOUSETYPE houseType = (HOUSETYPE) iter->newValue;

                HouseInfo& curHouseInfo = houseInfo[iter->slot];

                for(int i=0;i<curHouseInfo.HouseDropDown.GetNumEntries();i++) {
                    if(curHouseInfo.HouseDropDown.GetEntryIntData(i) == houseType) {
                        curHouseInfo.HouseDropDown.SetSelectedItem(i);
                        break;
                    }
                }
            } break;

            case ChangeEventList::ChangeEvent::EventType_ChangeTeam: {
                int newTeam = (int) iter->newValue;

                HouseInfo& curHouseInfo = houseInfo[iter->slot];

                for(int i=0;i<curHouseInfo.TeamDropDown.GetNumEntries();i++) {
                    if(curHouseInfo.TeamDropDown.GetEntryIntData(i) == newTeam) {
                        curHouseInfo.TeamDropDown.SetSelectedItem(i);
                        break;
                    }
                }
            } break;

            case ChangeEventList::ChangeEvent::EventType_ChangePlayer: {
                int newPlayer = (int) iter->newValue;

                HouseInfo& curHouseInfo = houseInfo[iter->slot/2];

                if(iter->slot % 2 == 0) {
                    for(int i=0;i<curHouseInfo.Player1DropDown.GetNumEntries();i++) {
                        if(curHouseInfo.Player1DropDown.GetEntryIntData(i) == newPlayer) {
                            curHouseInfo.Player1DropDown.SetSelectedItem(i);
                            break;
                        }
                    }
                } else {
                    for(int i=0;i<curHouseInfo.Player2DropDown.GetNumEntries();i++) {
                        if(curHouseInfo.Player2DropDown.GetEntryIntData(i) == newPlayer) {
                            curHouseInfo.Player2DropDown.SetSelectedItem(i);
                            break;
                        }
                    }
                }

                checkPlayerBoxes();
            } break;

            case ChangeEventList::ChangeEvent::EventType_SetHumanPlayer: {
                std::string name = iter->newStringValue;
                int slot = iter->slot;

                setPlayer2Slot(name, slot);

                checkPlayerBoxes();
            } break;

            default: {
            } break;
        }
    }

    if((pNetworkManager != NULL) && bServer) {
        ChangeEventList changeEventList2 = GetChangeEventList();

        pNetworkManager->sendChangeEventList(changeEventList2);
    }
}

ChangeEventList CustomGamePlayers::GetChangeEventList()
{
    ChangeEventList changeEventList;

    for(int i=0;i<NUM_HOUSES;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        int houseID = curHouseInfo.HouseDropDown.GetSelectedEntryIntData();
        int team = curHouseInfo.TeamDropDown.GetSelectedEntryIntData();
        int player1 = curHouseInfo.Player1DropDown.GetSelectedEntryIntData();
        int player2 = curHouseInfo.Player2DropDown.GetSelectedEntryIntData();

        changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangeHouse, i, houseID));
        changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangeTeam, i, team));

        if(player1 == PLAYER_HUMAN) {
            std::string playername = curHouseInfo.Player1DropDown.GetSelectedEntry();
            changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(2*i, playername));
        } else {
            changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangePlayer, 2*i, player1));
        }

        if(player2 == PLAYER_HUMAN) {
            std::string playername = curHouseInfo.Player2DropDown.GetSelectedEntry();
            changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(2*i+1, playername));
        } else {
            changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangePlayer, 2*i+1, player2));
        }
    }

    return changeEventList;
}

ChangeEventList CustomGamePlayers::GetChangeEventListForNewPlayer(std::string newPlayerName)
{
    ChangeEventList changeEventList = GetChangeEventList();

    // find a place for the new player

    int newPlayerSlot = INVALID;

    // look for an open left slot
    for(int i=0;i<numHouses;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];
        int player1 = curHouseInfo.Player1DropDown.GetSelectedEntryIntData();

        if(player1 == PLAYER_OPEN) {
            newPlayerSlot = 2*i;
            break;
        }
    }

    // if multiple players per house look for an open right slot
    if((newPlayerSlot == INVALID) && gameInitSettings.isMultiplePlayersPerHouse()) {
        for(int i=0;i<numHouses;i++) {
            HouseInfo& curHouseInfo = houseInfo[i];
            int player2 = curHouseInfo.Player2DropDown.GetSelectedEntryIntData();

            if(player2 == PLAYER_OPEN) {
                newPlayerSlot = 2*i + 1;
                break;
            }
        }
    }

    // as a fallback look for any non-human slot
    if(newPlayerSlot == INVALID) {
        for(int i=0;i<numHouses;i++) {
            HouseInfo& curHouseInfo = houseInfo[i];
            int player1 = curHouseInfo.Player1DropDown.GetSelectedEntryIntData();
            int player2 = curHouseInfo.Player2DropDown.GetSelectedEntryIntData();

            if(player1 != PLAYER_HUMAN) {
                newPlayerSlot = 2*i;
                break;
            } else if(gameInitSettings.isMultiplePlayersPerHouse() && (player2 != PLAYER_HUMAN)) {
                newPlayerSlot = 2*i + 1;
                break;
            }
        }
    }

    if(newPlayerSlot != INVALID) {
        setPlayer2Slot(newPlayerName, newPlayerSlot);

        ChangeEventList::ChangeEvent changeEvent(newPlayerSlot, newPlayerName);

        // add changeEvent to changeEventList for the new player
        changeEventList.changeEventList.push_back(changeEvent);

        // send it to all other connected peers
        ChangeEventList changeEventList2;
        changeEventList2.changeEventList.push_back(changeEvent);

        pNetworkManager->sendChangeEventList(changeEventList2);
    }

    return changeEventList;
}

void CustomGamePlayers::OnReceiveChatMessage(std::string name, std::string message) {
	addChatMessage(name, message);
}

void CustomGamePlayers::OnNext()
{
    // check if we have at least two houses on the map and if we have more than one team
    int numUsedHouses = 0;
    int numTeams = 0;
    for(int i=0;i<NUM_HOUSES;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        int currentPlayer1 = curHouseInfo.Player1DropDown.GetSelectedEntryIntData();
        int currentPlayer2 = curHouseInfo.Player2DropDown.GetSelectedEntryIntData();

        if((currentPlayer1 != PLAYER_OPEN && currentPlayer1 != PLAYER_CLOSED) || (currentPlayer2 != PLAYER_OPEN && currentPlayer2 != PLAYER_CLOSED)) {
            numUsedHouses++;

            int currentTeam = curHouseInfo.TeamDropDown.GetSelectedEntryIntData();
            bool bTeamFound = false;
            for(int j=0;j<i;j++) {
                int player1 = houseInfo[j].Player1DropDown.GetSelectedEntryIntData();
                int player2 = houseInfo[j].Player2DropDown.GetSelectedEntryIntData();

                if((player1 != PLAYER_OPEN && player1 != PLAYER_CLOSED) || (player2 != PLAYER_OPEN && player2 != PLAYER_CLOSED)) {

                    int team = houseInfo[j].TeamDropDown.GetSelectedEntryIntData();
                    if(currentTeam == team) {
                        bTeamFound = true;
                    }
                }
            }

            if(bTeamFound == false) {
                numTeams++;
            }
        }
    }

    if(numUsedHouses < 2) {
        // No game possible with only 1 house
        OpenWindow(MsgBox::Create(pTextManager->getLocalized("At least 2 houses must be controlled\nby a human player or an AI player!")));
    } else if(numTeams < 2) {
        // No game possible with only 1 team
        OpenWindow(MsgBox::Create(pTextManager->getLocalized("There must be at least two different teams!")));
    } else {
        // start game

        addAllPlayersToGameInitSettings();

        if(pNetworkManager != NULL) {
            unsigned int timeLeft = 5000;
            startGameTime = SDL_GetTicks() + timeLeft;
            pNetworkManager->sendStartGame(timeLeft);

            disableAllDropDownBoxes();
        } else {
            startSinglePlayerGame(gameInitSettings);

            quit(MENU_QUIT_GAME_FINISHED);
        }
    }
}

void CustomGamePlayers::addAllPlayersToGameInitSettings()
{
    gameInitSettings.clearHouseInfo();

    for(int i=0;i<NUM_HOUSES;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        int houseID = curHouseInfo.HouseDropDown.GetSelectedEntryIntData();
        int team = curHouseInfo.TeamDropDown.GetSelectedEntryIntData();
        int player1 = curHouseInfo.Player1DropDown.GetSelectedEntryIntData();
        std::string player1name = curHouseInfo.Player1DropDown.GetSelectedEntry();
        int player2 = curHouseInfo.Player2DropDown.GetSelectedEntryIntData();
        std::string player2name = curHouseInfo.Player2DropDown.GetSelectedEntry();

        GameInitSettings::HouseInfo newHouseInfo((HOUSETYPE) houseID, team);

        bool bAdded = false;
        bAdded |= addPlayerToHouseInfo(newHouseInfo, player1, player1name);
        bAdded |= addPlayerToHouseInfo(newHouseInfo, player2, player2name);

        if(bAdded == true) {
            gameInitSettings.addHouseInfo(newHouseInfo);
        }
    }
}

bool CustomGamePlayers::addPlayerToHouseInfo(GameInitSettings::HouseInfo& newHouseInfo, int player, std::string playername)
{
    std::string PlayerName;
    std::string PlayerClass;
    std::string PlayerClassParameter;

    HOUSETYPE houseID = newHouseInfo.HouseID;

    switch(player) {
        case PLAYER_HUMAN: {
            PlayerName = playername;
            PlayerClass = "HumanPlayer";
        } break;

        case PLAYER_AI_EASY: {
            PlayerName = (houseID == HOUSE_INVALID) ? "AI (easy)" : getHouseNameByNumber(houseID);
            PlayerClass = "AIPlayer";
            PlayerClassParameter = "0";
        } break;

        case PLAYER_AI_MEDIUM: {
            PlayerName = (houseID == HOUSE_INVALID) ? "AI (medium)" : getHouseNameByNumber(houseID);
            PlayerClass = "AIPlayer";
            PlayerClassParameter = "1";
        } break;

        case PLAYER_AI_DIFFICULT: {
            PlayerName = (houseID == HOUSE_INVALID) ? "AI (difficult)" : getHouseNameByNumber(houseID);
            PlayerClass = "AIPlayer";
            PlayerClassParameter = "2";
        } break;

        case PLAYER_OPEN:
        case PLAYER_CLOSED:
        default: {
            return false;
        }
    }

    newHouseInfo.addPlayerInfo(GameInitSettings::PlayerInfo(PlayerName, PlayerClass, PlayerClassParameter));
    return true;
}

void CustomGamePlayers::OnCancel()
{
    quit();
}

void CustomGamePlayers::OnSendChatMessage()
{
	std::string message = ChatTextBox.GetText();
	addChatMessage(settings.General.PlayerName, message);
	ChatTextBox.SetText("");

    pNetworkManager->sendChatMessage(message);

}

void CustomGamePlayers::addInfoMessage(std::string message) {
    std::string text = ChatTextView.GetText();
    if(text.length() > 0) {
        text += "\n";
    }
    text += "* " + message;
	ChatTextView.SetText(text);
	ChatTextView.scrollToEnd();
}

void CustomGamePlayers::addChatMessage(std::string name, std::string message)
{
    std::string text = ChatTextView.GetText();
    if(text.length() > 0) {
        text += "\n";
    }
    text += name + ": " + message;
	ChatTextView.SetText(text);
	ChatTextView.scrollToEnd();
}

void CustomGamePlayers::extractMapInfo(std::shared_ptr<INIFile>& pMap)
{
    int sizeX = 0;
    int sizeY = 0;

    if(pMap->hasKey("MAP","Seed")) {
        // old map format with seed value
        int mapscale = pMap->getIntValue("BASIC", "MapScale", -1);

        switch(mapscale) {
            case 0: {
                sizeX = 62;
                sizeY = 62;
            } break;

            case 1: {
                sizeX = 32;
                sizeY = 32;
            } break;

            case 2: {
                sizeX = 21;
                sizeY = 21;
            } break;

            default: {
                sizeX = 64;
                sizeY = 64;
            }
        }
    } else {
        // new map format with saved map
        sizeX = pMap->getIntValue("MAP","SizeX", 0);
        sizeY = pMap->getIntValue("MAP","SizeY", 0);
    }

    MapProperty_Size.SetText(stringify(sizeX) + " x " + stringify(sizeY));

    SDL_Surface* pMapSurface = GUIStyle::GetInstance().CreateButtonSurface(130,130,"", true, false);

    try {
        INIMapPreviewCreator mapPreviewCreator(pMap);

        SDL_Surface* pMinimap = mapPreviewCreator.createMinimapImageOfMap();
        SDL_Rect dest = { 1, 1, pMinimap->w, pMinimap->h};
        SDL_BlitSurface(pMinimap, NULL, pMapSurface, &dest);
        SDL_FreeSurface(pMinimap);
    } catch(...) {
        SDL_FreeSurface(pMapSurface);
        pMapSurface = GUIStyle::GetInstance().CreateButtonSurface(130,130,"Error", true, false);
        Button_Next.SetEnabled(false);
    }

    Minimap.SetSurface(pMapSurface, true);


    boundHousesOnMap.clear();
    if(pMap->hasSection("Atreides"))  boundHousesOnMap.push_back(HOUSE_ATREIDES);
    if(pMap->hasSection("Ordos"))     boundHousesOnMap.push_back(HOUSE_ORDOS);
    if(pMap->hasSection("Harkonnen")) boundHousesOnMap.push_back(HOUSE_HARKONNEN);
    if(pMap->hasSection("Fremen"))    boundHousesOnMap.push_back(HOUSE_FREMEN);
    if(pMap->hasSection("Mercenary")) boundHousesOnMap.push_back(HOUSE_MERCENARY);
    if(pMap->hasSection("Sardaukar")) boundHousesOnMap.push_back(HOUSE_SARDAUKAR);

    numHouses = boundHousesOnMap.size();
    if(pMap->hasSection("Player1"))   numHouses++;
    if(pMap->hasSection("Player2"))   numHouses++;
    if(pMap->hasSection("Player3"))   numHouses++;
    if(pMap->hasSection("Player4"))   numHouses++;
    if(pMap->hasSection("Player5"))   numHouses++;
    if(pMap->hasSection("Player6"))   numHouses++;

    MapProperty_Players.SetText(stringify(numHouses));

    std::string authors = pMap->getStringValue("BASIC","Author", "-");
    if(authors.size() > 11) {
        authors = authors.substr(0,9) + "...";
    }
    MapProperty_Authors.SetText(authors);


    MapProperty_License.SetText(pMap->getStringValue("BASIC","License", "-"));

}

void CustomGamePlayers::OnChangeHousesDropDownBoxes(bool bInteractive, int houseInfoNum) {
    if(bInteractive && houseInfoNum >= 0 && pNetworkManager != NULL) {
        int selectedHouseID = houseInfo[houseInfoNum].HouseDropDown.GetSelectedEntryIntData();

        ChangeEventList changeEventList;
        changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangeHouse, houseInfoNum, selectedHouseID));

        pNetworkManager->sendChangeEventList(changeEventList);
    }

    int numBoundHouses = boundHousesOnMap.size();
    int numUsedBoundHouses = 0;
    int numUsedRandomHouses = 0;
    for(int i=0;i<numHouses;i++) {
        int selectedHouseID = houseInfo[i].HouseDropDown.GetSelectedEntryIntData();

        if(selectedHouseID == HOUSE_INVALID) {
            numUsedRandomHouses++;
        } else {
            if(isBoundedHouseOnMap((HOUSETYPE) selectedHouseID)) {
                numUsedBoundHouses++;
            }
        }
    }


    for(int i=0;i<numHouses;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        int house = curHouseInfo.HouseDropDown.GetSelectedEntryIntData();

        if(houseInfoNum == -1 || houseInfoNum == i) {

            int color = ((house == HOUSE_INVALID) ? 234 : houseColor[house]) + 3;
            curHouseInfo.HouseLabel.SetTextColor(color);
            curHouseInfo.HouseDropDown.SetColor(color);
            curHouseInfo.TeamDropDown.SetColor(color);
            curHouseInfo.Player1Label.SetTextColor(color);
            curHouseInfo.Player1DropDown.SetColor(color);
            curHouseInfo.Player2Label.SetTextColor(color);
            curHouseInfo.Player2DropDown.SetColor(color);

            if(house == HOUSE_INVALID) {
                curHouseInfo.Player1ArrowLabel.SetSurface(pGFXManager->getUIGraphic(UI_CustomGamePlayersArrowNeutral), false);
            } else {
                curHouseInfo.Player1ArrowLabel.SetSurface(pGFXManager->getUIGraphic(UI_CustomGamePlayersArrow, house), false);
            }
        }

        if(gameInitSettings.getGameType() == GAMETYPE_LOAD_MULTIPLAYER) {
            // no house changes possible
            continue;
        }

        addToHouseDropDown(curHouseInfo.HouseDropDown, HOUSE_INVALID);

        for(int h=0;h<NUM_HOUSES;h++) {
            bool bAddHouse;

            bool bCheck;

            if((house == HOUSE_INVALID) || (isBoundedHouseOnMap((HOUSETYPE) house))) {
                if(numUsedBoundHouses + numUsedRandomHouses - 1 >= numBoundHouses) {
                    bCheck = true;
                } else {
                    bCheck = false;
                }
            } else {
                if(numUsedBoundHouses + numUsedRandomHouses >= numBoundHouses) {
                    bCheck = true;
                } else {
                    bCheck = false;
                }
            }

            if(bCheck == true) {
                bAddHouse = true;
                for(int j = 0; j < numHouses; j++) {
                    if(i != j) {
                        DropDownBox& tmpDropDown = houseInfo[j].HouseDropDown;
                        if(tmpDropDown.GetSelectedEntryIntData() == h) {
                            bAddHouse = false;
                            break;
                        }
                    }
                }
            } else {
                bAddHouse = (h == house);

                if(((house == HOUSE_INVALID) || (isBoundedHouseOnMap((HOUSETYPE) house))) && isBoundedHouseOnMap((HOUSETYPE) h)) {
                    // check if this entry is random or a bounded house but is needed for a bounded house
                    bAddHouse = true;
                    for(int j = 0; j < numHouses; j++) {
                        if(i != j) {
                            DropDownBox& tmpDropDown = houseInfo[j].HouseDropDown;
                            if(tmpDropDown.GetSelectedEntryIntData() == h) {
                                bAddHouse = false;
                                break;
                            }
                        }
                    }
                }
            }

            if(bAddHouse == true) {
                addToHouseDropDown(curHouseInfo.HouseDropDown, h);
            } else {
                removeFromHouseDropDown(curHouseInfo.HouseDropDown, h);
            }
        }
    }
}

void CustomGamePlayers::OnChangeTeamDropDownBoxes(bool bInteractive, int houseInfoNum) {

    if(bInteractive && houseInfoNum >= 0 && pNetworkManager != NULL) {
        int selectedTeam = houseInfo[houseInfoNum].TeamDropDown.GetSelectedEntryIntData();

        ChangeEventList changeEventList;
        changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangeTeam, houseInfoNum, selectedTeam));

        pNetworkManager->sendChangeEventList(changeEventList);
    }
}

void CustomGamePlayers::OnChangePlayerDropDownBoxes(bool bInteractive, int boxnum) {
    if(bInteractive && boxnum >= 0 && pNetworkManager != NULL) {
        DropDownBox& dropDownBox = (boxnum % 2 == 0) ? houseInfo[boxnum / 2].Player1DropDown : houseInfo[boxnum / 2].Player2DropDown;

        int selectedPlayer = dropDownBox.GetSelectedEntryIntData();

        ChangeEventList changeEventList;
        changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(ChangeEventList::ChangeEvent::EventType_ChangePlayer, boxnum, selectedPlayer));

        pNetworkManager->sendChangeEventList(changeEventList);
    }

    checkPlayerBoxes();
}

void CustomGamePlayers::OnClickPlayerDropDownBox(int boxnum) {
    DropDownBox& dropDownBox = (boxnum % 2 == 0) ? houseInfo[boxnum / 2].Player1DropDown : houseInfo[boxnum / 2].Player2DropDown;

    if(dropDownBox.GetSelectedEntryIntData() == PLAYER_CLOSED) {
        return;
    }

    setPlayer2Slot(settings.General.PlayerName, boxnum);

    if(boxnum >= 0 && pNetworkManager != NULL) {
        ChangeEventList changeEventList;
        changeEventList.changeEventList.push_back(ChangeEventList::ChangeEvent(boxnum, settings.General.PlayerName));

        pNetworkManager->sendChangeEventList(changeEventList);
    }
}

void CustomGamePlayers::OnPeerDisconnected(std::string playername, bool bHost, int cause) {
    if(bHost) {
        quit(cause);
    } else {
        for(int i=0;i<numHouses*2;i++) {
            DropDownBox& curDropDownBox = (i % 2 == 0) ? houseInfo[i / 2].Player1DropDown : houseInfo[i / 2].Player2DropDown;

            if(curDropDownBox.GetSelectedEntryIntData() == PLAYER_HUMAN && curDropDownBox.GetSelectedEntry() == playername) {

                curDropDownBox.ClearAllEntries();
                curDropDownBox.AddEntry(pTextManager->getLocalized("open"), PLAYER_OPEN);
                curDropDownBox.SetSelectedItem(0);

                if(gameInitSettings.getGameType() != GAMETYPE_LOAD_MULTIPLAYER) {
                    curDropDownBox.AddEntry(pTextManager->getLocalized("closed"), PLAYER_CLOSED);
                    curDropDownBox.AddEntry("AI (easy)", PLAYER_AI_EASY);
                    curDropDownBox.AddEntry("AI (medium)", PLAYER_AI_MEDIUM);
                    curDropDownBox.AddEntry("AI (difficult)", PLAYER_AI_DIFFICULT);
                }

                break;
            }
        }

        if(bServer == false) {
            for(int i=0;i<numHouses;i++) {
                bool bIsThisPlayer = false;
                if(houseInfo[i].Player1DropDown.GetSelectedEntryIntData() == PLAYER_HUMAN) {
                    if(houseInfo[i].Player1DropDown.GetSelectedEntry() == settings.General.PlayerName) {
                        bIsThisPlayer = true;
                    }
                }

                if(houseInfo[i].Player2DropDown.GetSelectedEntryIntData() == PLAYER_HUMAN) {
                    if(houseInfo[i].Player2DropDown.GetSelectedEntry() == settings.General.PlayerName) {
                        bIsThisPlayer = true;
                    }
                }

                houseInfo[i].Player1DropDown.SetEnabled(bIsThisPlayer);
                houseInfo[i].Player2DropDown.SetEnabled(bIsThisPlayer);
                houseInfo[i].HouseDropDown.SetEnabled(bIsThisPlayer);
                houseInfo[i].TeamDropDown.SetEnabled(bIsThisPlayer);
            }
        }

        checkPlayerBoxes();

        addInfoMessage(playername + " disconnected!");
    }
}

void CustomGamePlayers::OnStartGame(unsigned int timeLeft) {
    startGameTime = SDL_GetTicks() + timeLeft;
    disableAllDropDownBoxes();
}

void CustomGamePlayers::setPlayer2Slot(std::string playername, int slot) {
    DropDownBox& dropDownBox = (slot % 2 == 0) ? houseInfo[slot / 2].Player1DropDown : houseInfo[slot / 2].Player2DropDown;

    std::string oldPlayerName = "";
    int oldPlayerType = dropDownBox.GetSelectedEntryIntData();
    if(oldPlayerType == PLAYER_HUMAN) {
        oldPlayerName = dropDownBox.GetSelectedEntry();
    }

    for(int i=0;i<numHouses*2;i++) {
        DropDownBox& curDropDownBox = (i % 2 == 0) ? houseInfo[i / 2].Player1DropDown : houseInfo[i / 2].Player2DropDown;

        if(curDropDownBox.GetSelectedEntryIntData() == PLAYER_HUMAN && curDropDownBox.GetSelectedEntry() == playername) {

            if(oldPlayerType == PLAYER_HUMAN) {
                curDropDownBox.ClearAllEntries();
                curDropDownBox.AddEntry(oldPlayerName, PLAYER_HUMAN);
                curDropDownBox.SetSelectedItem(0);
            } else {
                curDropDownBox.ClearAllEntries();
                curDropDownBox.AddEntry(pTextManager->getLocalized("open"), PLAYER_OPEN);

                if(gameInitSettings.getGameType() != GAMETYPE_LOAD_MULTIPLAYER) {
                    curDropDownBox.AddEntry(pTextManager->getLocalized("closed"), PLAYER_CLOSED);
                    curDropDownBox.AddEntry("AI (easy)", PLAYER_AI_EASY);
                    curDropDownBox.AddEntry("AI (medium)", PLAYER_AI_MEDIUM);
                    curDropDownBox.AddEntry("AI (difficult)", PLAYER_AI_DIFFICULT);
                }

                for(int j = 0; j < curDropDownBox.GetNumEntries(); j++) {
                    if(curDropDownBox.GetEntryIntData(j) == oldPlayerType) {
                        curDropDownBox.SetSelectedItem(j);
                        break;
                    }
                }
            }

            break;
        }
    }

    dropDownBox.ClearAllEntries();
    dropDownBox.AddEntry(playername, PLAYER_HUMAN);
    dropDownBox.SetSelectedItem(0);

    if(bServer == false) {
        for(int i=0;i<numHouses;i++) {
            bool bIsThisPlayer = false;

            if(gameInitSettings.getGameType() != GAMETYPE_LOAD_MULTIPLAYER) {
                if(houseInfo[i].Player1DropDown.GetSelectedEntryIntData() == PLAYER_HUMAN) {
                    if(houseInfo[i].Player1DropDown.GetSelectedEntry() == settings.General.PlayerName) {
                        bIsThisPlayer = true;
                    }
                }

                if(houseInfo[i].Player2DropDown.GetSelectedEntryIntData() == PLAYER_HUMAN) {
                    if(houseInfo[i].Player2DropDown.GetSelectedEntry() == settings.General.PlayerName) {
                        bIsThisPlayer = true;
                    }
                }
            }

            houseInfo[i].Player1DropDown.SetEnabled(bIsThisPlayer);
            houseInfo[i].Player2DropDown.SetEnabled(bIsThisPlayer);
            houseInfo[i].HouseDropDown.SetEnabled(bIsThisPlayer);
            houseInfo[i].TeamDropDown.SetEnabled(bIsThisPlayer);
        }
    }

    checkPlayerBoxes();
}

void CustomGamePlayers::checkPlayerBoxes() {
    int numPlayers = 0;

    for(int i=0;i<numHouses;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        int player1 = curHouseInfo.Player1DropDown.GetSelectedEntryIntData();
        int player2 = curHouseInfo.Player2DropDown.GetSelectedEntryIntData();

        if(player1 != PLAYER_OPEN) {
            numPlayers++;
        }

        if(gameInitSettings.isMultiplePlayersPerHouse() && player2 != PLAYER_OPEN) {
            numPlayers++;
        }

        if((gameInitSettings.isMultiplePlayersPerHouse() == false) || (player1 == PLAYER_OPEN && player2 == PLAYER_OPEN) || (curHouseInfo.Player2DropDown.GetNumEntries() == 0)) {
            curHouseInfo.Player2DropDown.SetVisible(false);
            curHouseInfo.Player2DropDown.SetEnabled(false);
            curHouseInfo.Player2Label.SetVisible(false);
        } else {
            curHouseInfo.Player2DropDown.SetVisible(true);

            bool bEnableDropDown2 = bServer;
            if(bServer == false) {
                if(player1 == PLAYER_HUMAN) {
                    if(curHouseInfo.Player1DropDown.GetSelectedEntry() == settings.General.PlayerName) {
                        bEnableDropDown2 = true;
                    }
                }

                if(player2 == PLAYER_HUMAN) {
                    if(curHouseInfo.Player2DropDown.GetSelectedEntry() == settings.General.PlayerName) {
                        bEnableDropDown2 = true;
                    }
                }
            }

            if((gameInitSettings.getGameType() == GAMETYPE_LOAD_MULTIPLAYER) && (player2 != PLAYER_OPEN) && (player2 != PLAYER_HUMAN)) {
                curHouseInfo.Player2DropDown.SetEnabled(false);
            } else {
                curHouseInfo.Player2DropDown.SetEnabled(bEnableDropDown2);
            }
            curHouseInfo.Player2Label.SetVisible(true);
        }
    }

    if(pNetworkManager != NULL) {
        if(bServer) {
            pNetworkManager->updateServer(numPlayers);
        }
    }
}


void CustomGamePlayers::addToHouseDropDown(DropDownBox& houseDropDownBox, int house, bool bSelect) {

    if(houseDropDownBox.GetNumEntries() == 0) {
        if(house == HOUSE_INVALID) {
            houseDropDownBox.AddEntry(pTextManager->getLocalized("Random"), HOUSE_INVALID);
        } else {
            houseDropDownBox.AddEntry(getHouseNameByNumber((HOUSETYPE) house), house);
        }

        if(bSelect) {
            houseDropDownBox.SetSelectedItem(0);
        }
    } else {

        if(house == HOUSE_INVALID) {
            if(houseDropDownBox.GetEntryIntData(0) != HOUSE_INVALID) {
                houseDropDownBox.InsertEntry(0, pTextManager->getLocalized("Random"), HOUSE_INVALID);
            }

            if(bSelect) {
                houseDropDownBox.SetSelectedItem(houseDropDownBox.GetNumEntries()-1);
            }
        } else {

            int currentItemIndex = (houseDropDownBox.GetEntryIntData(0) == HOUSE_INVALID) ? 1 : 0;

            for(int h = 0; h < NUM_HOUSES; h++) {
                if(currentItemIndex < houseDropDownBox.GetNumEntries() && houseDropDownBox.GetEntryIntData(currentItemIndex) == h) {
                    if(h == house) {
                        if(bSelect) {
                            houseDropDownBox.SetSelectedItem(currentItemIndex);
                        }
                        break;
                    }

                    currentItemIndex++;
                } else {
                    if(h == house) {
                        houseDropDownBox.InsertEntry(currentItemIndex, getHouseNameByNumber((HOUSETYPE) h), h);

                        if(bSelect) {
                            houseDropDownBox.SetSelectedItem(currentItemIndex);
                        }
                        break;
                    }
                }


            }
        }
    }
}

void CustomGamePlayers::removeFromHouseDropDown(DropDownBox& houseDropDownBox, int house) {

    for(int i=0;i<houseDropDownBox.GetNumEntries(); i++) {
        if(houseDropDownBox.GetEntryIntData(i) == house) {
            houseDropDownBox.RemoveEntry(i);
            break;
        }
    }
}

bool CustomGamePlayers::isBoundedHouseOnMap(HOUSETYPE houseID) {
    std::list<HOUSETYPE>::const_iterator iter;
    for(iter = boundHousesOnMap.begin(); iter != boundHousesOnMap.end(); ++iter) {
        if(*iter == houseID) {
            return true;
        }
    }

    return false;
}

void CustomGamePlayers::disableAllDropDownBoxes() {
    for(int i=0;i<NUM_HOUSES;i++) {
        HouseInfo& curHouseInfo = houseInfo[i];

        curHouseInfo.HouseDropDown.SetEnabled(false);
        curHouseInfo.HouseDropDown.SetOnClickEnabled(false);
        curHouseInfo.TeamDropDown.SetEnabled(false);
        curHouseInfo.TeamDropDown.SetOnClickEnabled(false);
        curHouseInfo.Player1DropDown.SetEnabled(false);
        curHouseInfo.Player1DropDown.SetOnClickEnabled(false);
        curHouseInfo.Player2DropDown.SetEnabled(false);
        curHouseInfo.Player2DropDown.SetOnClickEnabled(false);
    }

    Button_Back.SetEnabled(false);
    Button_Next.SetEnabled(false);
}
