/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ObjectBase.h>

#include <globals.h>

#include <FileClasses/GFXManager.h>
#include <FileClasses/music/MusicPlayer.h>

#include <Game.h>
#include <House.h>
#include <SoundPlayer.h>
#include <Map.h>
#include <ScreenBorder.h>
#include <players/HumanPlayer.h>
#include <GUI/ObjectInterfaces/DefaultObjectInterface.h>

//structures
#include <structures/Barracks.h>
#include <structures/ConstructionYard.h>
#include <structures/GunTurret.h>
#include <structures/HeavyFactory.h>
#include <structures/HighTechFactory.h>
#include <structures/IX.h>
#include <structures/LightFactory.h>
#include <structures/Palace.h>
#include <structures/Radar.h>
#include <structures/Refinery.h>
#include <structures/RepairYard.h>
#include <structures/RocketTurret.h>
#include <structures/Silo.h>
#include <structures/StarPort.h>
#include <structures/Wall.h>
#include <structures/WindTrap.h>
#include <structures/WOR.h>

//units
#include <units/Carryall.h>
#include <units/Devastator.h>
#include <units/Deviator.h>
#include <units/FremenTrooper.h>
#include <units/Frigate.h>
#include <units/Harvester.h>
#include <units/Launcher.h>
#include <units/MCV.h>
#include <units/Ornithopter.h>
#include <units/Quad.h>
#include <units/Raider.h>
#include <units/Saboteur.h>
#include <units/SandWorm.h>
#include <units/SardaukarTrooper.h>
#include <units/SiegeTank.h>
#include <units/Soldier.h>
#include <units/SonicTank.h>
#include <units/Tank.h>
#include <units/Trike.h>
#include <units/Trooper.h>

ObjectBase::ObjectBase(House* newOwner) : owner(newOwner)
{
    ObjectBase::init();

	objectID = NONE;

    health = 0.0f;
    badlyDamaged = false;

	location.x = INVALID_POS;
	location.y = INVALID_POS;
    oldLocation.x = INVALID_POS;
	oldLocation.y = INVALID_POS;
	destination.x = INVALID_POS;
	destination.y = INVALID_POS;
	realX = 0.0f;
	realY = 0.0f;

    drawnAngle = 0;
	angle = (float) drawnAngle;

	active = false;
    respondable = true;
	selected = false;
	selectedByOtherPlayer = false;

	forced = false;
    setTarget(NULL);
	targetFriendly = false;
	attackMode = STOP;

	setVisible(VIS_ALL, false);
}

ObjectBase::ObjectBase(InputStream& stream)
{
    owner = currentGame->house[stream.readUint32()];

    ObjectBase::init();

    health = stream.readFloat();
    badlyDamaged = stream.readBool();

	location.x = stream.readSint32();
	location.y = stream.readSint32();
    oldLocation.x = stream.readSint32();
	oldLocation.y = stream.readSint32();
	destination.x = stream.readSint32();
	destination.y = stream.readSint32();
	realX = stream.readFloat();
	realY = stream.readFloat();

    angle = stream.readFloat();
    drawnAngle = stream.readSint8();

    active = stream.readBool();
    respondable = stream.readBool();

    if(currentGame->getGameInitSettings().getGameType() != GAMETYPE_CUSTOM_MULTIPLAYER) {
        selected = stream.readBool();
        selectedByOtherPlayer = stream.readBool();
    } else {
        selected = false;
        selectedByOtherPlayer = false;
    }

	forced = stream.readBool();
	target.load(stream);
	targetFriendly = stream.readBool();
	attackMode = (ATTACKMODE) stream.readUint32();

    stream.readBools(&visible[0], &visible[1], &visible[2], &visible[3], &visible[4], &visible[5]);
}

void ObjectBase::init() {
	itemID = ItemID_Invalid;

	aFlyingUnit = false;
	aGroundUnit = false;
	aStructure = false;
	aUnit = false;
	infantry = false;
	aBuilder = false;

	canAttackStuff = false;

	radius = TILESIZE/2;

	GraphicID = -1;
    graphic = NULL;
    numImagesX = 0;
    numImagesY = 0;

}

ObjectBase::~ObjectBase() {
}

void ObjectBase::save(OutputStream& stream) const
{
    stream.writeUint32(owner->getHouseID());

    stream.writeFloat(health);
    stream.writeBool(badlyDamaged);

	stream.writeSint32(location.x);
	stream.writeSint32(location.y);
    stream.writeSint32(oldLocation.x);
	stream.writeSint32(oldLocation.y);
	stream.writeSint32(destination.x);
	stream.writeSint32(destination.y);
	stream.writeFloat(realX);
	stream.writeFloat(realY);

    stream.writeFloat(angle);
    stream.writeSint8(drawnAngle);

	stream.writeBool(active);
	stream.writeBool(respondable);

	if(currentGame->getGameInitSettings().getGameType() != GAMETYPE_CUSTOM_MULTIPLAYER) {
        stream.writeBool(selected);
        stream.writeBool(selectedByOtherPlayer);
	}

	stream.writeBool(forced);
    target.save(stream);
	stream.writeBool(targetFriendly);
    stream.writeUint32(attackMode);

    stream.writeBools(visible[0], visible[1], visible[2], visible[3], visible[4], visible[5]);
}


/**
    Returns the center point of this object
    \return the center point in world coordinates
*/
Coord ObjectBase::getCenterPoint() const
{
    return Coord(lround(realX), lround(realY));
}

Coord ObjectBase::getClosestCenterPoint(const Coord& objectLocation) const
{
	return getCenterPoint();
}


int ObjectBase::getMaxHealth() const
{
    return currentGame->objectData.data[itemID].hitpoints;
}

void ObjectBase::handleDamage(int damage, Uint32 damagerID, House* damagerOwner)
{
    if(damage >= 0) {
        float newHealth = getHealth();

        newHealth -= damage;

        if(newHealth < 0.0f) {
            setHealth(0.0f);

            if(damagerOwner != NULL) {
                damagerOwner->informHasKilled(itemID);
            }
        } else {
            setHealth(newHealth);
        }
    }

    if(getOwner() == pLocalHouse) {
        musicPlayer->changeMusic(MUSIC_ATTACK);
    }

    getOwner()->noteDamageLocation(this, location);
}

void ObjectBase::HandleInterfaceEvent(SDL_Event* event)
{
	;
}

ObjectInterface* ObjectBase::GetInterfaceContainer() {
	return DefaultObjectInterface::Create(objectID);
}

void ObjectBase::removeFromSelectionLists()
{
    currentGame->getSelectedList().erase(getObjectID());
    currentGame->selectionChanged();
    currentGame->getSelectedByOtherPlayerList().erase(getObjectID());
    selected = false;

    for(int i=0; i < NUMSELECTEDLISTS; i++) {
        pLocalPlayer->getGroupList(i).erase(getObjectID());
    }
}

void ObjectBase::setDestination(int newX, int newY)
{
	if(currentGameMap->cellExists(newX, newY) || ((newX == INVALID_POS) && (newY == INVALID_POS))) {
		destination.x = newX;
		destination.y = newY;
	}
}

void ObjectBase::setHealth(float newHealth)
{
	if((newHealth >= 0.0f) && (newHealth <= getMaxHealth())) {
		health = newHealth;
        badlyDamaged = (health/(float)getMaxHealth() < HEAVILYDAMAGEDRATIO);
	}
}

void ObjectBase::setLocation(int xPos, int yPos)
{
	if((xPos == INVALID_POS) && (yPos == INVALID_POS)) {
		location.x = INVALID_POS;
		location.y = INVALID_POS;
	} else if (currentGameMap->cellExists(xPos, yPos))	{
		location.x = xPos;
		location.y = yPos;
		realX = location.x*TILESIZE;
		realY = location.y*TILESIZE;

		assignToMap(location);
	}
}

void ObjectBase::setObjectID(int newObjectID)
{
	if(newObjectID >= 0) {
		objectID = newObjectID;
	}
}

void ObjectBase::setVisible(int team, bool status)
{
	if(team == VIS_ALL) {
		for(int i = 0; i < MAX_PLAYERS; i++) {
			visible[i] = status;
		}
	} else if ((team >= 0) && (team < MAX_PLAYERS)) {
		visible[team] = status;
	}
}

void ObjectBase::setTarget(ObjectBase* newTarget)
{
	target.PointTo(newTarget);
	targetFriendly = (target && (target.getObjPointer()->getOwner()->getTeam() == owner->getTeam()) && (target.getObjPointer()->getItemID() != Unit_Sandworm));
}

void ObjectBase::unassignFromMap(const Coord& location)
{
	if (currentGameMap->cellExists(location))
		currentGameMap->getCell(location)->unassignObject(getObjectID());
}

bool ObjectBase::canAttack(const ObjectBase* object) const
{
	if ((object != NULL)
		&& (object->isAStructure()
			|| !object->isAFlyingUnit())
		&& ((object->getOwner()->getTeam() != owner->getTeam())
			|| object->getItemID() == Unit_Sandworm)
		&& object->isVisible(getOwner()->getTeam())) {
		return true;
	} else {
		return false;
	}
}

bool ObjectBase::isOnScreen() const
{
    Coord position = Coord((int) getRealX(), (int) getRealY());
    Coord size = Coord(graphic[currentZoomlevel]->w/numImagesX, graphic[currentZoomlevel]->h/numImagesY);

	if(screenborder->isInsideScreen(position,size) == true){
		return true;
	} else {
		return false;
	}
}

bool ObjectBase::isVisible(int team) const
{
	if ((team >= 0) && (team < MAX_PLAYERS))
		return visible[team];
	else
		return false;
}

bool ObjectBase::isVisible() const
{
    for(int i=0;i<MAX_PLAYERS;i++) {
        if(visible[i]) {
            return true;
        }
    }

    return false;
}

int ObjectBase::getHealthColor() const
{
	float healthPercent = (float)health/(float)getMaxHealth();

	if (healthPercent >= 0.7f) {
		return COLOR_LIGHTGREEN;
	} else if (healthPercent >= HEAVILYDAMAGEDRATIO) {
		return COLOR_YELLOW;
	} else {
		return COLOR_RED;
	}
}

Coord ObjectBase::getClosestPoint(const Coord& point) const
{
	return location;
}

StructureBase* ObjectBase::findClosestTargetStructure(ObjectBase* object)
{

	StructureBase	*closestStructure = NULL;
	float			closestDistance = INFINITY;

    RobustList<StructureBase*>::const_iterator iter;
    for(iter = structureList.begin(); iter != structureList.end(); ++iter) {
        StructureBase* tempStructure = *iter;

        if(object->canAttack(tempStructure)) {
			Coord closestPoint = tempStructure->getClosestPoint(object->getLocation());
			float structureDistance = blockDistance(object->getLocation(), closestPoint);

			if(tempStructure->getItemID() == Structure_Wall) {
					structureDistance += 20000000.0f; //so that walls are targeted very last
            }

            if(structureDistance < closestDistance)	{
                closestDistance = structureDistance;
                closestStructure = tempStructure;
            }
        }
	}

	return closestStructure;
}

UnitBase* ObjectBase::findClosestTargetUnit(ObjectBase* object)
{
	UnitBase	*closestUnit = NULL;
	float		closestDistance = INFINITY;

    RobustList<UnitBase*>::const_iterator iter;
    for(iter = unitList.begin(); iter != unitList.end(); ++iter) {
        UnitBase* tempUnit = *iter;

		if(object->canAttack(tempUnit)) {
			Coord closestPoint = tempUnit->getClosestPoint(object->getLocation());
			float unitDistance = blockDistance(object->getLocation(), closestPoint);

			if(unitDistance < closestDistance) {
                closestDistance = unitDistance;
				closestUnit = tempUnit;
            }
        }
	}

	return closestUnit;
}

ObjectBase* ObjectBase::findTarget()
{
 	ObjectBase	*tempTarget,
				*closestTarget = NULL;

	int	checkRange = 0;
	int	xPos = location.x;
	int	yPos = location.y;

	float closestDistance = INFINITY;

//searches for a target in an area like as shown below
//                     *****
//                   *********
//                  *****T*****
//                   *********
//                     *****

    switch(attackMode) {
        case GUARD: {
            checkRange = getWeaponRange();
        } break;

        case AREAGUARD: {
            checkRange = getGuardRange();
        } break;

        case AMBUSH: {
            checkRange = getViewRange();
        } break;

        case HUNT: {
            // check whole map
            StructureBase* pTargetStructure = findClosestTargetStructure(this);
            if(pTargetStructure != NULL) {
                return pTargetStructure;
            } else {
                return findClosestTargetUnit(this);
            }
        } break;

        case STOP:
        default: {
            return NULL;
        } break;
    }

	int xCheck = xPos - checkRange;

	if (xCheck < 0)
		xCheck = 0;

	int yCheck;

	while ((xCheck < currentGameMap->sizeX) && ((xCheck - xPos) <=  checkRange))
	{
		yCheck = (yPos - lookDist[abs(xCheck - xPos)]);

		if (yCheck < 0)
			yCheck = 0;

		while ((yCheck < currentGameMap->sizeY) && ((yCheck - yPos) <=  lookDist[abs(xCheck - xPos)]))
		{
			if(currentGameMap->cell[xCheck][yCheck].hasAnObject())
			{
				tempTarget = currentGameMap->cell[xCheck][yCheck].getObject();

				if (((tempTarget->getItemID() != Structure_Wall) || (closestTarget == NULL)) && canAttack(tempTarget))
				{
					float targetDistance = blockDistance(location, tempTarget->getLocation());
					if (targetDistance < closestDistance)
					{
						closestTarget = tempTarget;
						closestDistance = targetDistance;
					}
				}
			}

			yCheck++;
		}

		xCheck++;
		yCheck = yPos;
	}

	return closestTarget;
}

int ObjectBase::getViewRange() const
{
    return currentGame->objectData.data[itemID].viewrange;
}

int ObjectBase::getGuardRange() const {

    // small hack for first mission (otherwise all infantry units would attack the player)
    return std::min(DEFAULT_GUARDRANGE,
                    std::max(2*currentGame->objectData.data[itemID].viewrange,
                             (3*currentGame->objectData.data[itemID].weaponrange)/2 ));
}

int ObjectBase::getWeaponRange() const
{
    return currentGame->objectData.data[itemID].weaponrange;
}

int ObjectBase::getRadius() const
{
    return currentGame->objectData.data[itemID].radius;
}

ObjectBase* ObjectBase::createObject(int ItemID, House* Owner, Uint32 ObjectID)
{
	ObjectBase* newObject = NULL;
	switch (ItemID)
	{
		case Structure_Barracks:			newObject = new Barracks(Owner); break;
		case Structure_ConstructionYard:	newObject = new ConstructionYard(Owner); break;
		case Structure_GunTurret:			newObject = new GunTurret(Owner); break;
		case Structure_HeavyFactory:		newObject = new HeavyFactory(Owner); break;
		case Structure_HighTechFactory:		newObject = new HighTechFactory(Owner); break;
		case Structure_IX:					newObject = new IX(Owner); break;
		case Structure_LightFactory:		newObject = new LightFactory(Owner); break;
		case Structure_Palace:				newObject = new Palace(Owner); break;
		case Structure_Radar:				newObject = new Radar(Owner); break;
		case Structure_Refinery:			newObject = new Refinery(Owner); break;
		case Structure_RepairYard:			newObject = new RepairYard(Owner); break;
		case Structure_RocketTurret:		newObject = new RocketTurret(Owner); break;
		case Structure_Silo:				newObject = new Silo(Owner); break;
		case Structure_StarPort:			newObject = new StarPort(Owner); break;
		case Structure_Wall:				newObject = new Wall(Owner); break;
		case Structure_WindTrap:			newObject = new WindTrap(Owner); break;
		case Structure_WOR:					newObject = new WOR(Owner); break;

		case Unit_Carryall:					newObject = new Carryall(Owner); break;
		case Unit_Devastator:				newObject = new Devastator(Owner); break;
		case Unit_Deviator:					newObject = new Deviator(Owner); break;
		case Unit_Frigate:					newObject = new Frigate(Owner); break;
		case Unit_Harvester:				newObject = new Harvester(Owner); break;
		case Unit_Soldier:					newObject = new Soldier(Owner); break;
		case Unit_Launcher:					newObject = new Launcher(Owner); break;
		case Unit_MCV:						newObject = new MCV(Owner); break;
		case Unit_Ornithopter:				newObject = new Ornithopter(Owner); break;
		case Unit_Quad:						newObject = new Quad(Owner); break;
		case Unit_Saboteur:					newObject = new Saboteur(Owner); break;
		case Unit_Sandworm:					newObject = new Sandworm(Owner); break;
		case Unit_SiegeTank:				newObject = new SiegeTank(Owner); break;
		case Unit_SonicTank:				newObject = new SonicTank(Owner); break;
		case Unit_Tank:						newObject = new Tank(Owner); break;
		case Unit_Trike:					newObject = new Trike(Owner); break;
		case Unit_Raider:					newObject = new Raider(Owner); break;
		case Unit_Trooper:					newObject = new Trooper(Owner); break;
		case Unit_SardaukarTrooper:			newObject = new SardaukarTrooper(Owner); break;
		case Unit_FremenTrooper:			newObject = new FremenTrooper(Owner); break;

		default:							newObject = NULL;
											fprintf(stderr,"ObjectBase::createObject(): %d is no valid ItemID!\n",ItemID);
											break;
	}

	if(newObject == NULL)
		return NULL;

	if(ObjectID == NONE) {
		ObjectID = currentGame->getObjectManager().AddObject(newObject);
		newObject->setObjectID(ObjectID);
	} else {
		newObject->setObjectID(ObjectID);
	}

	return newObject;
}

ObjectBase* ObjectBase::loadObject(InputStream& stream, int ItemID, Uint32 ObjectID)
{
	ObjectBase* newObject = NULL;
	switch (ItemID)
	{
		case Structure_Barracks:			newObject = new Barracks(stream); break;
		case Structure_ConstructionYard:	newObject = new ConstructionYard(stream); break;
		case Structure_GunTurret:			newObject = new GunTurret(stream); break;
		case Structure_HeavyFactory:		newObject = new HeavyFactory(stream); break;
		case Structure_HighTechFactory:		newObject = new HighTechFactory(stream); break;
		case Structure_IX:					newObject = new IX(stream); break;
		case Structure_LightFactory:		newObject = new LightFactory(stream); break;
		case Structure_Palace:				newObject = new Palace(stream); break;
		case Structure_Radar:				newObject = new Radar(stream); break;
		case Structure_Refinery:			newObject = new Refinery(stream); break;
		case Structure_RepairYard:			newObject = new RepairYard(stream); break;
		case Structure_RocketTurret:		newObject = new RocketTurret(stream); break;
		case Structure_Silo:				newObject = new Silo(stream); break;
		case Structure_StarPort:			newObject = new StarPort(stream); break;
		case Structure_Wall:				newObject = new Wall(stream); break;
		case Structure_WindTrap:			newObject = new WindTrap(stream); break;
		case Structure_WOR:					newObject = new WOR(stream); break;

		case Unit_Carryall:					newObject = new Carryall(stream); break;
		case Unit_Devastator:				newObject = new Devastator(stream); break;
		case Unit_Deviator:					newObject = new Deviator(stream); break;
		case Unit_Frigate:					newObject = new Frigate(stream); break;
		case Unit_Harvester:				newObject = new Harvester(stream); break;
		case Unit_Soldier:					newObject = new Soldier(stream); break;
		case Unit_Launcher:					newObject = new Launcher(stream); break;
		case Unit_MCV:						newObject = new MCV(stream); break;
		case Unit_Ornithopter:				newObject = new Ornithopter(stream); break;
		case Unit_Quad:						newObject = new Quad(stream); break;
		case Unit_Saboteur:					newObject = new Saboteur(stream); break;
		case Unit_Sandworm:					newObject = new Sandworm(stream); break;
		case Unit_SiegeTank:				newObject = new SiegeTank(stream); break;
		case Unit_SonicTank:				newObject = new SonicTank(stream); break;
		case Unit_Tank:						newObject = new Tank(stream); break;
		case Unit_Trike:					newObject = new Trike(stream); break;
		case Unit_Raider:					newObject = new Raider(stream); break;
		case Unit_Trooper:					newObject = new Trooper(stream); break;
		case Unit_SardaukarTrooper:			newObject = new SardaukarTrooper(stream); break;
		case Unit_FremenTrooper:			newObject = new FremenTrooper(stream); break;

		default:							newObject = NULL;
											fprintf(stderr,"ObjectBase::loadObject(): %d is no valid ItemID!\n",ItemID);
											break;
	}

	if(newObject == NULL)
		return NULL;

	newObject->setObjectID(ObjectID);

	return newObject;
}

bool ObjectBase::targetInWeaponRange() const {
    Coord coord = (target.getObjPointer())->getClosestPoint(location);
    float dist = blockDistance(location,coord);
    return ( dist <= currentGame->objectData.data[itemID].weaponrange);
}

