/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <misc/draw_util.h>

#include <globals.h>

#include <stdexcept>


Uint32 getPixel(SDL_Surface *surface, int x, int y)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + (y * surface->pitch) + (x * bpp);

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        throw std::runtime_error("getPixel(): Invalid bpp value!");
    }
}


void putPixel(SDL_Surface *surface, int x, int y, Uint32 color)
{
	if(x >= 0 && x < surface->w && y >=0 && y < surface->h) {
		int bpp = surface->format->BytesPerPixel;
		/* Here p is the address to the pixel want to set */
		Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

		switch(bpp) {
		case 1:
			*p = color;
			break;

		case 2:
			*(Uint16 *)p = color;
			break;

		case 3:
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
				p[0] = (color>> 16) & 0xff;
				p[1] = (color>> 8) & 0xff;
				p[2] = color& 0xff;
			} else {
				p[0] = color& 0xff;
				p[1] = (color>> 8) & 0xff;
				p[2] = (color>> 16) & 0xff;
			}
			break;

		case 4:
			*(Uint32 *)p = color;
			break;
		}
	}
}


void drawHLineNoLock(SDL_Surface *surface, int x1, int y, int x2, Uint32 color)
{
	int	min = x1;
	int	max = x2;

	if(min > max) {
		int temp = max;
		max = min;
		min = temp;
	}

	for(int i = min; i <= max; i++) {
		putPixel(surface, i, y, color);
	}
}


void drawVLineNoLock(SDL_Surface *surface, int x, int y1, int y2, Uint32 color)
{
	int	min = y1;
	int	max = y2;

	if(min > max) {
		int temp = max;
		max = min;
		min = temp;
	}

	for(int i = min; i <= max; i++) {
		putPixel(surface, x, i, color);
	}
}


void drawHLine(SDL_Surface *surface, int x1, int y, int x2, Uint32 color)
{
	if(!SDL_MUSTLOCK(surface) || (SDL_LockSurface(surface) == 0)) {
		drawHLineNoLock(surface, x1, y, x2, color);

		if(SDL_MUSTLOCK(surface)) {
			SDL_UnlockSurface(surface);
		}
	}
}


void drawVLine(SDL_Surface *surface, int x, int y1, int y2, Uint32 color)
{
	if(!SDL_MUSTLOCK(surface) || (SDL_LockSurface(surface) == 0)) {
		drawVLineNoLock(surface, x, y1, y2, color);

		if (SDL_MUSTLOCK(surface)) {
			SDL_UnlockSurface(surface);
		}
	}
}


void drawRect(SDL_Surface *surface, int x1, int y1, int x2, int y2, Uint32 color)
{
	if(!SDL_MUSTLOCK(surface) || (SDL_LockSurface(surface) == 0)) {
		int	min = x1;
		int max = x2;

		if(min > max) {
			int temp = max;
			max = min;
			min = temp;
		}

		for(int i = min; i <= max; i++) {
			putPixel(surface, i, y1, color);
			putPixel(surface, i, y2, color);
		}

		min = y1+1;
		max = y2;
		if(min > max) {
			int temp = max;
			max = min;
			min = temp;
		}

		for(int j = min; j < max; j++) {
			putPixel(surface, x1, j, color);
			putPixel(surface, x2, j, color);
		}

		if(SDL_MUSTLOCK(surface)) {
			SDL_UnlockSurface(surface);
		}
	}
}

void replaceColor(SDL_Surface *surface, Uint32 oldColor, Uint32 newColor)
{
	if(!SDL_MUSTLOCK(surface) || (SDL_LockSurface(surface) == 0)) {
        for(int y = 0; y < surface->h; y++) {
            for(int x = 0; x < surface->w; x++) {
                if(getPixel(surface, x, y) == oldColor) {
                    putPixel(surface, x, y, newColor);
                }
            }
        }

        if(SDL_MUSTLOCK(surface)) {
			SDL_UnlockSurface(surface);
		}
	}
}


SDL_Surface* copySurface(SDL_Surface* inSurface)
{
	//return SDL_DisplayFormat(inSurface);
	return SDL_ConvertSurface(inSurface, inSurface->format, inSurface->flags);
}


SDL_Surface* doubleSurface(SDL_Surface* inputPic, bool bFreeInputPic) {
	if(inputPic == NULL) {
	    throw std::invalid_argument("doubleSurface(): inputPic == NULL!");
	}

	SDL_Surface *returnPic;

	// create new picture surface
	if((returnPic = SDL_CreateRGBSurface(SDL_HWSURFACE,inputPic->w * 2,inputPic->h * 2,8,0,0,0,0))== NULL) {
	    if(bFreeInputPic) SDL_FreeSurface(inputPic);
		throw std::runtime_error("doubleSurface(): Cannot create new Picture!");
	}

	SDL_SetColors(returnPic, inputPic->format->palette->colors, 0, inputPic->format->palette->ncolors);
	SDL_LockSurface(returnPic);
	SDL_LockSurface(inputPic);

	//Now we can copy pixel by pixel
	for(int y = 0; y < inputPic->h;y++) {
		for(int x = 0; x < inputPic->w; x++) {
			char val = *( ((char*) (inputPic->pixels)) + y*inputPic->pitch + x);
			*( ((char*) (returnPic->pixels)) + 2*y*returnPic->pitch + 2*x) = val;
			*( ((char*) (returnPic->pixels)) + 2*y*returnPic->pitch + 2*x+1) = val;
			*( ((char*) (returnPic->pixels)) + (2*y+1)*returnPic->pitch + 2*x) = val;
			*( ((char*) (returnPic->pixels)) + (2*y+1)*returnPic->pitch + 2*x+1) = val;
		}
	}

	SDL_UnlockSurface(inputPic);
	SDL_UnlockSurface(returnPic);

    if(bFreeInputPic) SDL_FreeSurface(inputPic);

	return returnPic;
}


SDL_Surface* scale2xSurface(SDL_Surface* src, int tilesX, int tilesY, bool freeSrcSurface) {
	if(src == NULL) {
		return NULL;
	}

	int srcWidth = src->w;
	int srcHeight = src->h;

	SDL_Surface* dest = SDL_CreateRGBSurface(src->flags, srcWidth*2, srcHeight*2, 8, 0, 0, 0, 0);
	if(dest == NULL) {
		fprintf(stderr,"Test\n");
		return NULL;
	}
	palette.applyToSurface(dest);

	Uint8* srcPixels = (Uint8*) src->pixels;
	Uint8* destPixels = (Uint8*) dest->pixels;

	int tileWidth = srcWidth / tilesX;
	int tileHeight = srcHeight / tilesY;

	for(int j=0;j<tilesY;++j) {
		for(int i=0;i<tilesX;++i) {



			/*

			Scale center pixel E into 4 new pixels

				Source			  Dest
			+---+---+---+
			| A | B | C |		+--+--+
			+---+---+---+		|E0|E1|
			| D | E | F |	->	+--+--+
			+---+---+---+		|E2|E3|
			| G | H | I |		+--+--+
			+---+---+---+

			*/

			for(int y = 0; y < tileHeight; ++y) {
				for(int x = 0; x < tileWidth; ++x) {
					Uint8 E = *( srcPixels + (j*tileHeight+y)*src->pitch + (i*tileWidth+x) );
					Uint8 B = *( srcPixels + (j*tileHeight+std::max(0,y-1))*src->pitch + (i*tileWidth+x) );
					Uint8 H = *( srcPixels + (j*tileHeight+std::min(tileHeight-1,y+1))*src->pitch + (i*tileWidth+x) );
					Uint8 D = *( srcPixels + (j*tileHeight+y)*src->pitch + (i*tileWidth+std::max(0,x-1)) );
					Uint8 F = *( srcPixels + (j*tileHeight+y)*src->pitch + (i*tileWidth+std::min(tileWidth-1,x+1)) );

					Uint8 E0, E1, E2, E3;

					if(B != H && D != F) {
						E0 = (D == B) ? D : E;
						E1 = (B == F) ? F : E;
						E2 = (D == H) ? D : E;
						E3 = (H == F) ? F : E;
					} else {
						E0 = E;
						E1 = E;
						E2 = E;
						E3 = E;
					}

					*(destPixels + (j*tileHeight+y)*2*dest->pitch + (i*tileWidth+x)*2) = E0;
					*(destPixels + (j*tileHeight+y)*2*dest->pitch + (i*tileWidth+x)*2 + 1) = E1;
					*(destPixels + ((j*tileHeight+y)*2+1)*dest->pitch + (i*tileWidth+x)*2) = E2;
					*(destPixels + ((j*tileHeight+y)*2+1)*dest->pitch + (i*tileWidth+x)*2 + 1) = E3;
				}
			}

		}
	}

	if(freeSrcSurface) {
		SDL_FreeSurface(src);
	}

	return dest;
}


SDL_Surface* scale3xSurface(SDL_Surface* src, int tilesX, int tilesY, bool freeSrcSurface) {
	if(src == NULL) {
		return NULL;
	}

	int srcWidth = src->w;
	int srcHeight = src->h;

	SDL_Surface* dest = SDL_CreateRGBSurface(src->flags, srcWidth*3, srcHeight*3, 8, 0, 0, 0, 0);
	if(dest == NULL) {
		fprintf(stderr,"Test\n");
		return NULL;
	}
	palette.applyToSurface(dest);

	Uint8* srcPixels = (Uint8*) src->pixels;
	Uint8* destPixels = (Uint8*) dest->pixels;

	int tileWidth = srcWidth / tilesX;
	int tileHeight = srcHeight / tilesY;

	for(int j=0;j<tilesY;++j) {
		for(int i=0;i<tilesX;++i) {



			/*

			Scale center pixel E into 4 new pixels

				Source			   Dest
			+---+---+---+		+--+--+--+
			| A | B | C |		|E0|E1|E2|
			+---+---+---+		+--+--+--+
			| D | E | F |	->	|E3|E4|E5|
			+---+---+---+		+--+--+--+
			| G | H | I |		|E6|E7|E8|
			+---+---+---+		+--+--+--+

			*/

			for(int y = 0; y < tileHeight; ++y) {
				for(int x = 0; x < tileWidth; ++x) {
					Uint8 A = *( srcPixels + (j*tileHeight+std::max(0,y-1))*src->pitch + (i*tileWidth+std::max(0,x-1)) );
					Uint8 B = *( srcPixels + (j*tileHeight+std::max(0,y-1))*src->pitch + (i*tileWidth+x) );
					Uint8 C = *( srcPixels + (j*tileHeight+std::max(0,y-1))*src->pitch + (i*tileWidth+std::min(tileWidth-1,x+1)) );
					Uint8 D = *( srcPixels + (j*tileHeight+y)*src->pitch + (i*tileWidth+std::max(0,x-1)) );
					Uint8 E = *( srcPixels + (j*tileHeight+y)*src->pitch + (i*tileWidth+x) );
					Uint8 F = *( srcPixels + (j*tileHeight+y)*src->pitch + (i*tileWidth+std::min(tileWidth-1,x+1)) );
					Uint8 G = *( srcPixels + (j*tileHeight+std::min(tileHeight-1,y+1))*src->pitch + (i*tileWidth+std::max(0,x-1)) );
					Uint8 H = *( srcPixels + (j*tileHeight+std::min(tileHeight-1,y+1))*src->pitch + (i*tileWidth+x) );
					Uint8 I = *( srcPixels + (j*tileHeight+std::min(tileHeight-1,y+1))*src->pitch + (i*tileWidth+std::min(tileWidth-1,x+1)) );

					Uint8 E0, E1, E2, E3, E4, E5, E6, E7, E8;

					if(B != H && D != F) {
						E0 = (D == B) ? D : E;
						E1 = (((D == B) && (E != C)) || ((B == F) && (E != A))) ? B : E;
						E2 = (B == F) ? F : E;
						E3 = (((D == B && E != G)) || ((D == H) && (E != A))) ? D : E;
						E4 = E;
						E5 = (((B == F) && (E != I)) || ((H == F) && (E != C))) ? F : E;
						E6 = (D == H) ? D : E;
						E7 = (((D == H) && (E != I)) || ((H == F) && (E != G))) ? H : E;
						E8 = (H == F) ? F : E;
					} else {
						E0 = E;
						E1 = E;
						E2 = E;
						E3 = E;
						E4 = E;
						E5 = E;
						E6 = E;
						E7 = E;
						E8 = E;
					}

					*(destPixels + (j*tileHeight+y)*3*dest->pitch + (i*tileWidth+x)*3) = E0;
					*(destPixels + (j*tileHeight+y)*3*dest->pitch + (i*tileWidth+x)*3 + 1) = E1;
					*(destPixels + (j*tileHeight+y)*3*dest->pitch + (i*tileWidth+x)*3 + 2) = E2;
					*(destPixels + ((j*tileHeight+y)*3+1)*dest->pitch + (i*tileWidth+x)*3) = E3;
					*(destPixels + ((j*tileHeight+y)*3+1)*dest->pitch + (i*tileWidth+x)*3 + 1) = E4;
					*(destPixels + ((j*tileHeight+y)*3+1)*dest->pitch + (i*tileWidth+x)*3 + 2) = E5;
					*(destPixels + ((j*tileHeight+y)*3+2)*dest->pitch + (i*tileWidth+x)*3) = E6;
					*(destPixels + ((j*tileHeight+y)*3+2)*dest->pitch + (i*tileWidth+x)*3 + 1) = E7;
					*(destPixels + ((j*tileHeight+y)*3+2)*dest->pitch + (i*tileWidth+x)*3 + 2) = E8;
				}
			}

		}
	}

	if(freeSrcSurface) {
		SDL_FreeSurface(src);
	}

	return dest;
}


SDL_Surface* scaleSurface(SDL_Surface *surf, double ratio)
{
	SDL_Surface *scaled = SDL_CreateRGBSurface(SDL_HWSURFACE, (int) (surf->w * ratio),(int) (surf->h * ratio),32,0,0,0,0);

	SDL_LockSurface(scaled);
	SDL_LockSurface(surf);

	int X2 = (int)(surf->w * ratio);
	int Y2 = (int)(surf->h * ratio);

	for(int x = 0;x < X2;++x)
		for(int y = 0;y < Y2;++y)
			putPixel(scaled,x,y,getPixel(surf,(int) (x/ratio), (int) (y/ratio)));

	SDL_UnlockSurface(scaled);
	SDL_UnlockSurface(surf);

	return scaled;
}


SDL_Surface* getSubPicture(SDL_Surface* Pic, unsigned int left, unsigned int top,
											unsigned int width, unsigned int height) {
	if(Pic == NULL) {
	    throw std::invalid_argument("getSubPicture(): Pic == NULL!");
	}

	if(((int) (left+width) > Pic->w) || ((int) (top+height) > Pic->h)) {
		throw std::invalid_argument("getSubPicture(): left+width > Pic->w || top+height > Pic->h!");
	}

	SDL_Surface *returnPic;

	// create new picture surface
	if((returnPic = SDL_CreateRGBSurface(SDL_HWSURFACE,width,height,8,0,0,0,0))== NULL) {
		throw std::runtime_error("getSubPicture(): Cannot create new Picture!");
	}

	SDL_SetColors(returnPic, Pic->format->palette->colors, 0, Pic->format->palette->ncolors);

	SDL_Rect srcRect = {left,top,width,height};
	SDL_BlitSurface(Pic,&srcRect,returnPic,NULL);

	return returnPic;
}


SDL_Surface* rotateSurfaceLeft(SDL_Surface* inputPic, bool bFreeInputPic) {
	if(inputPic == NULL) {
		throw std::invalid_argument("rotateSurfaceLeft(): inputPic == NULL!");
	}

	SDL_Surface *returnPic;

	// create new picture surface
	if((returnPic = SDL_CreateRGBSurface(SDL_HWSURFACE,inputPic->h,inputPic->w,8,0,0,0,0))== NULL) {
	    if(bFreeInputPic) SDL_FreeSurface(inputPic);
		throw std::runtime_error("rotateSurfaceLeft(): Cannot create new Picture!");
	}

	SDL_SetColors(returnPic, inputPic->format->palette->colors, 0, inputPic->format->palette->ncolors);
	SDL_LockSurface(returnPic);
	SDL_LockSurface(inputPic);

	//Now we can copy pixel by pixel
	for(int y = 0; y < inputPic->h;y++) {
		for(int x = 0; x < inputPic->w; x++) {
			char val = *( ((char*) (inputPic->pixels)) + y*inputPic->pitch + x);
			*( ((char*) (returnPic->pixels)) + (returnPic->h - x - 1)*returnPic->pitch + y) = val;
		}
	}

	SDL_UnlockSurface(inputPic);
	SDL_UnlockSurface(returnPic);

    if(bFreeInputPic) SDL_FreeSurface(inputPic);

	return returnPic;
}


SDL_Surface* rotateSurfaceRight(SDL_Surface* inputPic, bool bFreeInputPic) {
	if(inputPic == NULL) {
		throw std::invalid_argument("rotateSurfaceRight(): inputPic == NULL!");
	}

	SDL_Surface *returnPic;

	// create new picture surface
	if((returnPic = SDL_CreateRGBSurface(SDL_HWSURFACE,inputPic->h,inputPic->w,8,0,0,0,0))== NULL) {
	    if(bFreeInputPic) SDL_FreeSurface(inputPic);
		throw std::runtime_error("rotateSurfaceRight(): Cannot create new Picture!");
	}

	SDL_SetColors(returnPic, inputPic->format->palette->colors, 0, inputPic->format->palette->ncolors);
	SDL_LockSurface(returnPic);
	SDL_LockSurface(inputPic);

	//Now we can copy pixel by pixel
	for(int y = 0; y < inputPic->h;y++) {
		for(int x = 0; x < inputPic->w; x++) {
			char val = *( ((char*) (inputPic->pixels)) + y*inputPic->pitch + x);
			*( ((char*) (returnPic->pixels)) + x*returnPic->pitch + (returnPic->w - y - 1)) = val;
		}
	}

	SDL_UnlockSurface(inputPic);
	SDL_UnlockSurface(returnPic);

    if(bFreeInputPic) SDL_FreeSurface(inputPic);

	return returnPic;
}


SDL_Surface* flipHSurface(SDL_Surface* inputPic, bool bFreeInputPic) {
	if(inputPic == NULL) {
		throw std::invalid_argument("flipHSurface(): inputPic == NULL!");
	}

	SDL_Surface *returnPic;

	// create new picture surface
	if((returnPic = SDL_CreateRGBSurface(SDL_HWSURFACE,inputPic->w,inputPic->h,8,0,0,0,0))== NULL) {
	    if(bFreeInputPic) SDL_FreeSurface(inputPic);
		throw std::runtime_error("flipHSurface(): Cannot create new Picture!");
	}

	SDL_SetColors(returnPic, inputPic->format->palette->colors, 0, inputPic->format->palette->ncolors);
	SDL_LockSurface(returnPic);
	SDL_LockSurface(inputPic);

	//Now we can copy pixel by pixel
	for(int y = 0; y < inputPic->h;y++) {
		for(int x = 0; x < inputPic->w; x++) {
			char val = *( ((char*) (inputPic->pixels)) + y*inputPic->pitch + x);
			*( ((char*) (returnPic->pixels)) + (returnPic->h - y - 1)*returnPic->pitch + x) = val;
		}
	}

	SDL_UnlockSurface(inputPic);
	SDL_UnlockSurface(returnPic);

    if(bFreeInputPic) SDL_FreeSurface(inputPic);

	return returnPic;
}


SDL_Surface* flipVSurface(SDL_Surface* inputPic, bool bFreeInputPic) {
	if(inputPic == NULL) {
		throw std::invalid_argument("flipVSurface(): inputPic == NULL!");
	}

	SDL_Surface *returnPic;

	// create new picture surface
	if((returnPic = SDL_CreateRGBSurface(SDL_HWSURFACE,inputPic->w,inputPic->h,8,0,0,0,0))== NULL) {
	    if(bFreeInputPic) SDL_FreeSurface(inputPic);
		throw std::runtime_error("flipVSurface(): Cannot create new Picture!");
	}

	SDL_SetColors(returnPic, inputPic->format->palette->colors, 0, inputPic->format->palette->ncolors);
	SDL_LockSurface(returnPic);
	SDL_LockSurface(inputPic);

	//Now we can copy pixel by pixel
	for(int y = 0; y < inputPic->h;y++) {
		for(int x = 0; x < inputPic->w; x++) {
			char val = *( ((char*) (inputPic->pixels)) + y*inputPic->pitch + x);
			*( ((char*) (returnPic->pixels)) + y*returnPic->pitch + (inputPic->w - x - 1)) = val;
		}
	}

	SDL_UnlockSurface(inputPic);
	SDL_UnlockSurface(returnPic);

    if(bFreeInputPic) SDL_FreeSurface(inputPic);

	return returnPic;
}


SDL_Surface* createShadowSurface(SDL_Surface* source)
{
    if(source == NULL) {
	    throw std::invalid_argument("createShadowSurface(): source == NULL!");
	}

	SDL_Surface *retPic;

	if((retPic = SDL_ConvertSurface(source,source->format,SDL_HWSURFACE)) == NULL) {
	    throw std::runtime_error("createShadowSurface(): Cannot copy image!");
	}

	if(SDL_LockSurface(retPic) != 0) {
	    SDL_FreeSurface(retPic);
	    throw std::runtime_error("createShadowSurface(): Cannot lock image!");
	}

	for(int i = 0; i < retPic->w; i++) {
		for(int j = 0; j < retPic->h; j++) {
			Uint8* pixel = &((Uint8*)retPic->pixels)[j * retPic->pitch + i];
			if(*pixel != 0) {
                *pixel = 12;
			}
		}
	}
	SDL_UnlockSurface(retPic);

	return retPic;
}


SDL_Surface* mapSurfaceColorRange(SDL_Surface* source, int srcColor, int destColor)
{
	SDL_Surface *retPic;

	if((retPic = SDL_ConvertSurface(source,source->format,SDL_HWSURFACE)) == NULL) {
	    throw std::runtime_error("mapSurfaceColorRange(): Cannot copy image!");
	}

	if(SDL_LockSurface(retPic) != 0) {
	    SDL_FreeSurface(retPic);
	    throw std::runtime_error("mapSurfaceColorRange(): Cannot lock image!");
	}

	for(int i = 0; i < retPic->w; i++) {
		for(int j = 0; j < retPic->h; j++) {
			Uint8* pixel = &((Uint8*)retPic->pixels)[j * retPic->pitch + i];
			if ((*pixel >= srcColor) && (*pixel < srcColor + 7))
				*pixel = *pixel - srcColor + destColor;
		}
	}
	SDL_UnlockSurface(retPic);

	return retPic;
}
