/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <units/Carryall.h>

#include <globals.h>

#include <FileClasses/GFXManager.h>
#include <House.h>
#include <Map.h>
#include <Game.h>

#include <structures/RepairYard.h>
#include <structures/Refinery.h>
#include <structures/ConstructionYard.h>
#include <units/Harvester.h>

#include <misc/strictmath.h>

Carryall::Carryall(House* newOwner) : AirUnit(newOwner)
{
    Carryall::init();

    setHealth(getMaxHealth());

	booked = false;
    idle = true;
	firstRun = true;
    owned = true;

    aDropOfferer = false;
    droppedOffCargo = false;
    respondable = false;

    currentMaxSpeed = 2.0f;

	curFlyPoint = 0;
	for(int i=0; i < 8; i++) {
		flyPoints[i].x = INVALID_POS;
		flyPoints[i].y = INVALID_POS;
	}
	constYardPoint.x = INVALID_POS;
	constYardPoint.y = INVALID_POS;
}

Carryall::Carryall(InputStream& stream) : AirUnit(stream)
{
    Carryall::init();

	PickedUpUnitList = stream.readUint32List();
	if(!PickedUpUnitList.empty()) {
		drawnFrame = 1;
	}

    stream.readBools(&booked, &idle, &firstRun, &owned, &aDropOfferer, &droppedOffCargo);

	currentMaxSpeed = stream.readFloat();

	curFlyPoint = stream.readUint8();
	for(int i=0; i < 8; i++) {
		flyPoints[i].x = stream.readSint32();
		flyPoints[i].y = stream.readSint32();
	}
	constYardPoint.x = stream.readSint32();
	constYardPoint.y = stream.readSint32();
}

void Carryall::init()
{
	itemID = Unit_Carryall;
	owner->incrementUnits(itemID);

	canAttackStuff = false;

	GraphicID = ObjPic_Carryall;
	graphic = pGFXManager->getObjPic(GraphicID,getOwner()->getHouseID());
	shadowGraphic = pGFXManager->getObjPic(ObjPic_CarryallShadow,getOwner()->getHouseID());

	numImagesX = NUM_ANGLES;
	numImagesY = 2;
}

Carryall::~Carryall()
{
	;
}

void Carryall::save(OutputStream& stream) const
{
	AirUnit::save(stream);

	stream.writeUint32List(PickedUpUnitList);

    stream.writeBools(booked, idle, firstRun, owned, aDropOfferer, droppedOffCargo);

	stream.writeFloat(currentMaxSpeed);

	stream.writeUint8(curFlyPoint);
	for(int i=0; i < 8; i++) {
		stream.writeSint32(flyPoints[i].x);
		stream.writeSint32(flyPoints[i].y);
	}
	stream.writeSint32(constYardPoint.x);
	stream.writeSint32(constYardPoint.y);
}

bool Carryall::update() {

    if(AirUnit::update() == false) {
        return false;
    }

    float dist = distance_from(location.x*TILESIZE + TILESIZE/2, location.y*TILESIZE + TILESIZE/2,
                                destination.x*TILESIZE + TILESIZE/2, destination.y*TILESIZE + TILESIZE/2);

    if((target || hasCargo()) && dist < 256.0f) {
        currentMaxSpeed = (((2.0f - currentGame->objectData.data[itemID].maxspeed)/256.0f) * (256.0f - dist)) + currentGame->objectData.data[itemID].maxspeed;
        setSpeeds();
    } else {
        currentMaxSpeed = std::min(currentMaxSpeed + 0.2f, currentGame->objectData.data[itemID].maxspeed);
        setSpeeds();
    }

	// check if this carryall has to be removed because it has just brought something
	// to the map (e.g. new harvester)
	if (active)	{
		if(aDropOfferer && droppedOffCargo && (hasCargo() == false)
            && (    (location.x == 0) || (location.x == currentGameMap->sizeX-1)
                    || (location.y == 0) || (location.y == currentGameMap->sizeY-1))
            && !moving)	{

            setVisible(VIS_ALL, false);
            destroy();
            return false;
		}
	}
	return true;
}

float Carryall::getMaxSpeed() const {
    return currentMaxSpeed;
}

void Carryall::deploy(const Coord& newLocation) {
	AirUnit::deploy(newLocation);

	respondable = false;
}

void Carryall::checkPos()
{
	AirUnit::checkPos();

	if (active)	{
		if (hasCargo() && (location == destination) && (distance_from(realX, realY, destination.x * TILESIZE + (TILESIZE/2), destination.y * TILESIZE + (TILESIZE/2)) < TILESIZE/8) ) {

		    // drop up to 3 infantry units at once or one other unit
            int droppedUnits = 0;
            do {
                Uint32 UnitID = PickedUpUnitList.front();
                UnitBase* unit = (UnitBase*) (currentGame->getObjectManager().getObject(UnitID));

                if((unit->isInfantry() == false) && (droppedUnits > 0)) {
                    // we already dropped infantry and this is no infantry
                    // => do not drop this here
                    break;
                }

                deployUnit(UnitID);
                droppedUnits++;

                if((unit->isInfantry() == false)) {
                    // we dropped a non infantry unit
                    // => do not drop another unit
                    break;
                }
            } while(hasCargo() && (droppedUnits < 3));

            if(PickedUpUnitList.empty() == false) {
                // find next place to drop
                for(int i=8;i<18;i++) {
                    int r = currentGame->RandomGen.rand(3,i/2);
                    float angle = 2.0f*PI*currentGame->RandomGen.randFloat();

                    Coord dropCoord = location + Coord( (int) (r*strictmath::sin(angle)), (int) (-r*strictmath::cos(angle)));
                    if(currentGameMap->cellExists(dropCoord) && currentGameMap->getCell(dropCoord)->hasAGroundObject() == false) {
                        setDestination(dropCoord);
                        break;
                    }
                }
            } else {
                setTarget(NULL);
                setDestination(guardPoint);

                idle = true;
            }
		} else if((isBooked() == false) && idle && !firstRun) {
			//fly around const yard
			Coord point = this->getClosestPoint(location);

			if(point == guardPoint) {
				//arrived at point, move to next
				curFlyPoint++;

				if(curFlyPoint >= 8) {
					curFlyPoint = 0;
				}

                int looped = 0;
				while(!(currentGameMap->cellExists(flyPoints[curFlyPoint].x, flyPoints[curFlyPoint].y)) && looped <= 2) {
					curFlyPoint++;

					if(curFlyPoint >= 8) {
						curFlyPoint = 0;
						looped++;
					}
				}

				setGuardPoint(flyPoints[curFlyPoint]);
				setDestination(guardPoint);
			}
		} else if(firstRun && owned) {
			findConstYard();
			setGuardPoint(constYardPoint);
			setDestination(guardPoint);
			firstRun = false;
		}
	}

}

void Carryall::deployUnit(Uint32 UnitID)
{
	bool found = false;

	std::list<Uint32>::iterator iter;
	for(iter = PickedUpUnitList.begin() ; iter != PickedUpUnitList.end(); ++iter) {
		if(*iter == UnitID) {
			found = true;
			break;
		}
	}

	if(found == false) {
        return;
	}

	PickedUpUnitList.remove(UnitID);

	UnitBase* unit = (UnitBase*) (currentGame->getObjectManager().getObject(UnitID));

	if(unit == NULL)
		return;

	if (found) {
	    currentMaxSpeed = 0.0f;
	    setSpeeds();

	    if (currentGameMap->getCell(location)->hasANonInfantryGroundObject()) {
			ObjectBase* object = currentGameMap->getCell(location)->getNonInfantryGroundObject();
			if (object->getOwner() == getOwner()) {
				if (object->getItemID() == Structure_RepairYard) {
					if (((RepairYard*)object)->isFree()) {
						unit->setTarget(object);
						unit->setGettingRepaired();
						unit = NULL;
					} else {
					    // carryall has booked this repair yard but now will not go there => unbook
						((RepairYard*)object)->unBook();

						// unit is still going to repair yard but was unbooked from repair yard at pickup => book now
						((RepairYard*)object)->book();
					}
				} else if ((object->getItemID() == Structure_Refinery) && (unit->getItemID() == Unit_Harvester)) {
					if (((Refinery*)object)->isFree()) {
						((Harvester*)unit)->setTarget(object);
						((Harvester*)unit)->setReturned();
						unit = NULL;
						goingToRepairYard = false;
					}
				}
			}
		}

		if(unit != NULL) {
			unit->setAngle(drawnAngle);
			Coord deployPos = currentGameMap->findDeploySpot(unit, location);
			unit->deploy(deployPos);
		}

		if (PickedUpUnitList.empty())
		{
			if(!aDropOfferer) {
				booked = false;
                idle = true;
			}
			droppedOffCargo = true;
			drawnFrame = 0;
			recalculatePathTimer = 0;
		}
	}
}

void Carryall::destroy()
{
    // destroy cargo
	std::list<Uint32>::const_iterator iter;
	for(iter = PickedUpUnitList.begin() ; iter != PickedUpUnitList.end(); ++iter) {
		UnitBase* unit = (UnitBase*) (currentGame->getObjectManager().getObject(*iter));
		if(unit != NULL) {
			unit->destroy();
		}
	}
	PickedUpUnitList.clear();

	// place wreck
    if(isVisible() && currentGameMap->cellExists(location)) {
        Tile* pTile = currentGameMap->getCell(location);
        pTile->assignDeadUnit(DeadUnit_Carrall, owner->getHouseID(), Coord((Sint32) realX, (Sint32) realY));
    }

	AirUnit::destroy();
}

void Carryall::engageTarget()
{
    if ((target.getObjPointer() == NULL)
		|| (!target.getObjPointer()->isVisible(getOwner()->getTeam()))
		|| (target.getObjPointer()->isAGroundUnit()
			&& !((GroundUnit*)target.getObjPointer())->isAwaitingPickup())
		|| (target.getObjPointer()->getOwner()->getTeam() != owner->getTeam())) {

		setTarget(NULL);

		if (!hasCargo()) {
			booked = false;
			idle = true;
			setDestination(guardPoint);
			nextSpotFound = false;
		}

	} else {
		Coord targetLocation;
		if(target.getObjPointer()->getItemID() == Structure_Refinery) {
		    targetLocation = target.getObjPointer()->getLocation() + Coord(2,0);
		} else {
            targetLocation = target.getObjPointer()->getClosestPoint(location);
		}
		Coord realLocation, realDestination;
		realLocation.x = lround(realX);
		realLocation.y = lround(realY);
		realDestination.x = targetLocation.x * TILESIZE + TILESIZE/2;
		realDestination.y = targetLocation.y * TILESIZE + TILESIZE/2;

		targetAngle = lround((float)NUM_ANGLES*dest_angle(location, destination)/256.0f);
		if (targetAngle == 8) {
			targetAngle = 0;
		}

		targetDistance = distance_from(realLocation, realDestination);

		if (targetDistance <= TILESIZE/8) {
			if (target.getObjPointer()->isAUnit()) {
				targetAngle = ((GroundUnit*)target.getObjPointer())->getAngle();
			}

			//if (drawnAngle == targetAngle) {
			    if(hasCargo()) {
			        if(target.getObjPointer()->isAStructure()) {
                        while(PickedUpUnitList.begin() != PickedUpUnitList.end()) {
                            deployUnit(*(PickedUpUnitList.begin()) );
                        }

                        setTarget(NULL);
                        setDestination(guardPoint);
                    }
			    } else {
                    pickupTarget();
			    }
			//}
		} else {
			setDestination(targetLocation);
		}
	}
}

void Carryall::giveCargo(UnitBase* newUnit)
{
	if(newUnit == NULL)
		return;

	booked = true;
	PickedUpUnitList.push_back(newUnit->getObjectID());

	newUnit->setPickedUp(this);

	if (getItemID() != Unit_Frigate)
		drawnFrame = 1;

	droppedOffCargo = false;
}

void Carryall::DoSetDeployStructure(Uint32 TargetObjectID) {
	if (target.getObjPointer() != NULL && target.getObjPointer()->isAStructure()) {
		if (goingToRepairYard) {
			((RepairYard*)target.getObjPointer())->unBook();
			goingToRepairYard = false;
			target.PointTo(NONE);
		} else if (target.getObjPointer()->getItemID() == Structure_Refinery) {
			((Refinery*)target.getObjPointer())->unBook();
			target.PointTo(NONE);
		}
	}

	pathList.clear();

	if(TargetObjectID == NONE) {
		std::list<Uint32>::iterator iter;
		for(iter = PickedUpUnitList.begin() ; iter != PickedUpUnitList.end(); ++iter) {
			deployUnit(*iter);
		}
		PickedUpUnitList.clear();
	} else {
		ObjectBase* pTarget = currentGame->getObjectManager().getObject(TargetObjectID);

		if(pTarget == NULL) {
            return;
		}

		if (pTarget->isAStructure())	{
			int targetID = pTarget->getItemID();
			if (targetID == Structure_RepairYard) {
				((RepairYard*) pTarget)->book();
				goingToRepairYard = true;
			} else if (targetID == Structure_Refinery) {
				((Refinery*) pTarget)->book();
			}
		}
	}
}

void Carryall::pickupTarget()
{
    currentMaxSpeed = 0.0f;
    setSpeeds();

	if(target.getObjPointer()->isAGroundUnit()) {

		if (target.getObjPointer()->hasATarget()
			|| ( ((GroundUnit*)target.getObjPointer())->getGuardPoint() != target.getObjPointer()->getLocation())
			|| ((GroundUnit*)target.getObjPointer())->isBadlyDamaged())	{

			if(((GroundUnit*)target.getObjPointer())->isBadlyDamaged() || (target.getObjPointer()->hasATarget() == false && target.getObjPointer()->getItemID() != Unit_Harvester))	{
				((GroundUnit*)target.getObjPointer())->DoRepair();
			}

			ObjectBase* newTarget = target.getObjPointer()->hasATarget() ? target.getObjPointer()->getTarget() : NULL;

			PickedUpUnitList.push_back(target.getObjectID());
			((GroundUnit*)target.getObjPointer())->setPickedUp(this);

			drawnFrame = 1;
			booked = true;

            if(newTarget && ((newTarget->getItemID() == Structure_Refinery)
                              || (newTarget->getItemID() == Structure_RepairYard)))
            {
                setTarget(newTarget);
                if(newTarget->getItemID() == Structure_Refinery) {
                    setDestination(target.getObjPointer()->getLocation() + Coord(2,0));
                } else {
                    setDestination(target.getObjPointer()->getClosestPoint(location));
                }
            } else if (((GroundUnit*)target.getObjPointer())->getDestination().x != INVALID_POS) {
                setDestination(((GroundUnit*)target.getObjPointer())->getDestination());
            }
            clearPath();
            recalculatePathTimer = 0;
		} else {
			((GroundUnit*)target.getObjPointer())->setAwaitingPickup(false);
			setTarget(NULL);
		}
	} else {
        // get unit from structure
        ObjectBase* pObject = target.getObjPointer();
        if(pObject->getItemID() == Structure_Refinery) {
            // get harvester
            ((Refinery*) pObject)->deployHarvester(this);
        } else if(pObject->getItemID() == Structure_RepairYard) {
            // get repaired unit
            ((RepairYard*) pObject)->deployRepairUnit(this);
        }
	}
}

void Carryall::setTarget(ObjectBase* newTarget)
{
	if (target.getObjPointer() != NULL
		&& targetFriendly
		&& target.getObjPointer()->isAGroundUnit()
		&& (((GroundUnit*)target.getObjPointer())->getCarrier() == this))
	{
		((GroundUnit*)target.getObjPointer())->bookCarrier(NULL);
	}

	UnitBase::setTarget(newTarget);

	if (target && targetFriendly && target.getObjPointer()->isAGroundUnit())
		((GroundUnit*)target.getObjPointer())->setAwaitingPickup(true);

	booked = target;
}

void Carryall::targeting()
{
	if (target)
		engageTarget();
}

void Carryall::turn()
{
	if(!moving) {
		int wantedAngle;

		if(target && (targetDistance <= currentGame->objectData.data[itemID].weaponrange)) {
			wantedAngle = targetAngle;
		} else {
			wantedAngle = nextSpotAngle;
		}

		if(wantedAngle != -1) {
			if(justStoppedMoving) {
				angle = wantedAngle;
				drawnAngle = lround(angle);
			} else {
				float	angleLeft = 0.0f;
                float   angleRight = 0.0f;

				if(angle > wantedAngle)	{
					angleRight = angle - wantedAngle;
					angleLeft = std::abs(8.0f-angle)+wantedAngle;
				} else if (angle < wantedAngle)	{
					angleRight = std::abs(8.0f-wantedAngle) + angle;
					angleLeft = wantedAngle - angle;
				}

				if (angleLeft <= angleRight) {
					turnLeft();
				} else {
					turnRight();
				}
			}
		}
	}
}
bool Carryall::hasCargo()
{
	return (PickedUpUnitList.size() > 0);
}

void Carryall::findConstYard()
{
    float	closestYardDistance = 1000000.0f;
    ConstructionYard* bestYard = NULL;

    RobustList<StructureBase*>::const_iterator iter;
    for(iter = structureList.begin(); iter != structureList.end(); ++iter) {
        StructureBase* tempStructure = *iter;

        if((tempStructure->getItemID() == Structure_ConstructionYard) && (tempStructure->getOwner() == owner)) {
            ConstructionYard* tempYard = ((ConstructionYard*) tempStructure);
            Coord closestPoint = tempYard->getClosestPoint(location);
            float tempDistance = distance_from(location, closestPoint);

            if(tempDistance < closestYardDistance) {
                closestYardDistance = tempDistance;
                bestYard = tempYard;
            }
        }
    }

    if(bestYard) {
        constYardPoint = bestYard->getClosestPoint(location);
    } else {
        constYardPoint = guardPoint;
    }

    static const Coord circles[][8] = {
                                        { Coord(-2,-6), Coord(3,-6), Coord(7,-2), Coord(7,3), Coord(3,7), Coord(-2,7), Coord(-6,3), Coord(-6,-2) },
                                        { Coord(-2,-6), Coord(-6,-2), Coord(-6,3), Coord(-2,7), Coord(3,7), Coord(7,3), Coord(7,-2), Coord(3,-6) },
                                        { Coord(-3,-8), Coord(4,-8), Coord(9,-3), Coord(9,4), Coord(4,9), Coord(-3,9), Coord(-8,4), Coord(-8,-3) },
                                        { Coord(-3,-8), Coord(-8,-3), Coord(-8,4), Coord(-3,9), Coord(4,9), Coord(9,4), Coord(9,-3), Coord(4,-8) }
                                      };

    const Coord* pUsedCircle = circles[currentGame->RandomGen.rand(0,3)];


    for(int i=0;i<8; i++) {
        flyPoints[i] = constYardPoint + pUsedCircle[i];
    }
}
