/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <units/InfantryBase.h>

#include <globals.h>

#include <House.h>
#include <Game.h>
#include <Map.h>
#include <SoundPlayer.h>
#include <ScreenBorder.h>

#include <players/HumanPlayer.h>

#include <structures/StructureBase.h>
#include <structures/Refinery.h>
#include <structures/RepairYard.h>
#include <units/Harvester.h>

InfantryBase::InfantryBase(House* newOwner) : GroundUnit(newOwner)
{
    InfantryBase::init();

    setHealth(getMaxHealth());

    cellPosition = INVALID_POS;
    oldCellPosition = INVALID_POS;
}

InfantryBase::InfantryBase(InputStream& stream) : GroundUnit(stream)
{
    InfantryBase::init();

	cellPosition = stream.readSint8();
	oldCellPosition = stream.readSint8();
}

void InfantryBase::init()
{
	infantry = true;

	walkFrame = 0;
}

InfantryBase::~InfantryBase()
{
}


void InfantryBase::save(OutputStream& stream) const
{
	GroundUnit::save(stream);

	stream.writeSint8(cellPosition);
	stream.writeSint8(oldCellPosition);
}

void InfantryBase::HandleCaptureClick(int xPos, int yPos) {
	if(respondable && ((getItemID() == Unit_FremenTrooper) || (getItemID() == Unit_SardaukarTrooper) || (getItemID() == Unit_Soldier) || (getItemID() == Unit_Trooper))) {
		if (currentGameMap->cellExists(xPos, yPos)) {
			if (currentGameMap->cell[xPos][yPos].hasAnObject()) {
				// capture structure
				ObjectBase* tempTarget = currentGameMap->cell[xPos][yPos].getObject();

				currentGame->GetCommandManager().addCommand(Command(pLocalPlayer->getPlayerID(), CMD_INFANTRY_CAPTURE,objectID,tempTarget->getObjectID()));
			}
		}
	}

}

void InfantryBase::DoCaptureObject(Uint32 TargetStructureID) {
	StructureBase* pStructure = dynamic_cast<StructureBase*>(currentGame->getObjectManager().getObject(TargetStructureID));

	if((pStructure == NULL) || (pStructure->canBeCaptured() == false) || (pStructure->getOwner()->getTeam() == getOwner()->getTeam())) {
	    // does not exist anymore, cannot be captured or is a friendly building
        return;
	}

	DoAttackObject(pStructure);
	DoSetAttackMode(CAPTURE);
}

void InfantryBase::assignToMap(const Coord& pos)
{
	if(currentGameMap->cellExists(pos)) {
		oldCellPosition = cellPosition;
		cellPosition = currentGameMap->cell[pos.x][pos.y].assignInfantry(getObjectID());
	}
}

void InfantryBase::blitToScreen()
{
    int imageW = graphic[currentZoomlevel]->w/numImagesX;
    int imageH = graphic[currentZoomlevel]->h/numImagesY;

	SDL_Rect dest = {   screenborder->world2screenX(realX) - imageW/2,
                        screenborder->world2screenY(realY) - imageH/2,
                        imageW, imageH };

    int temp = drawnAngle;
    if(temp == UP) {
        temp = 1;
    } else if (temp == DOWN) {
        temp = 3;
    } else if (temp == LEFTUP || temp == LEFTDOWN || temp == LEFT) {
        temp = 2;
    } else {
        //RIGHT
        temp = 0;
    }

    SDL_Rect source = { temp*imageW, (walkFrame/10 == 3) ? imageH : walkFrame/10*imageH, imageW, imageH };

    SDL_BlitSurface(graphic[currentZoomlevel], &source, screen, &dest);
}

bool InfantryBase::canPass(int xPos, int yPos) const
{
	bool passable = false;
	if (currentGameMap->cellExists(xPos, yPos))
	{
		Tile* cell = currentGameMap->getCell(xPos, yPos);
		if (!cell->hasAGroundObject())
		{
			if (cell->getType() != Terrain_Mountain)
				passable = true;
			else
			{
				/* if this unit is infantry so can climb, and cell can take more infantry */
				if (cell->infantryNotFull())
					passable = true;
			}
		}
		else
		{
			ObjectBase *object = cell->getGroundObject();

			if ((object != NULL) && (object->getObjectID() == target.getObjectID())
				&& object->isAStructure()
				&& (object->getOwner()->getTeam() != owner->getTeam())
				&& object->isVisible(getOwner()->getTeam()))
			{
				passable = true;
			} else {
				passable = (!cell->hasANonInfantryGroundObject()
							&& (cell->infantryNotFull()
							&& (cell->getInfantryTeam() == getOwner()->getTeam())));
			}
		}
	}
	return passable;
}

void InfantryBase::checkPos()
{
	if(moving) {
		if(++walkFrame > 39) {
			walkFrame = 0;
		}
	}

	if(justStoppedMoving) {
		walkFrame = 0;

		if(currentGameMap->getCell(location)->isSpiceBloom()) {
		    setHealth(0.0f);
			currentGameMap->getCell(location)->triggerSpiceBloom(getOwner());
		} else if(currentGameMap->getCell(location)->isSpecialBloom()){
            currentGameMap->getCell(location)->triggerSpecialBloom(getOwner());
		}

        //check to see if close enough to blow up target
        if(target.getObjPointer() != NULL
            && target.getObjPointer()->isAStructure()
            && (getOwner()->getTeam() != target.getObjPointer()->getOwner()->getTeam()))
        {
            Coord	closestPoint;

            closestPoint = target.getObjPointer()->getClosestPoint(location);

            if(blockDistance(location, closestPoint) <= 0.5f) {
                StructureBase* pCapturedStructure = target.getStructurePointer();
                if(pCapturedStructure->getHealthColor() == COLOR_RED) {
                    House* pOwner = pCapturedStructure->getOwner();
                    int targetID = pCapturedStructure->getItemID();
                    int posX = pCapturedStructure->getX();
                    int posY = pCapturedStructure->getY();
                    int origHouse = pCapturedStructure->getOriginalHouse();
                    int oldHealth = pCapturedStructure->getHealth();
                    bool isSelected = pCapturedStructure->isSelected();
                    bool isSelectedByOtherPlayer = pCapturedStructure->isSelectedByOtherPlayer();

                    float capturedSpice = 0.0f;

                    UnitBase* pContainedUnit = NULL;

                    if(pCapturedStructure->getItemID() == Structure_Silo) {
                        capturedSpice = currentGame->objectData.data[Structure_Silo].capacity * (pOwner->getStoredCredits() / pOwner->getCapacity());
                    } else if(pCapturedStructure->getItemID() == Structure_Refinery) {
                        capturedSpice = currentGame->objectData.data[Structure_Silo].capacity * (pOwner->getStoredCredits() / pOwner->getCapacity());
                        Refinery* pRefinery = dynamic_cast<Refinery*>(pCapturedStructure);
                        if(pRefinery->isFree() == false) {
                            pContainedUnit = pRefinery->getHarvester();
                        }
                    } else if(pCapturedStructure->getItemID() == Structure_RepairYard) {
                        RepairYard* pRepairYard = dynamic_cast<RepairYard*>(pCapturedStructure);
                        if(pRepairYard->isFree() == false) {
                            pContainedUnit = pRepairYard->getRepairUnit();
                        }
                    }

                    Uint32 containedUnitID = NONE;
                    float containedUnitHealth = 0.0f;
                    float containedHarvesterSpice = 0.0f;
                    if(pContainedUnit != NULL) {
                        containedUnitID = pContainedUnit->getItemID();
                        containedUnitHealth = pContainedUnit->getHealth();
                        if(containedUnitID == Unit_Harvester) {
                            containedHarvesterSpice = dynamic_cast<Harvester*>(pContainedUnit)->getAmountOfSpice();
                        }

                        // will be destroyed by the captured structure
                        pContainedUnit = NULL;
                    }

                    // remove all other infantry units capturing this building
                    Coord capturedStructureLocation = pCapturedStructure->getLocation();
                    for(int i = capturedStructureLocation.x; i < capturedStructureLocation.x + pCapturedStructure->getStructureSizeX(); i++) {
                        for(int j = capturedStructureLocation.y; j < capturedStructureLocation.y + pCapturedStructure->getStructureSizeY(); j++) {
                            std::list<Uint32> infantryList = currentGameMap->getCell(i,j)->getInfantryList();
                            std::list<Uint32>::iterator iter;
                            for(iter = infantryList.begin(); iter != infantryList.end(); ++iter) {
                                if(*iter != getObjectID()) {
                                    ObjectBase* pObject = currentGame->getObjectManager().getObject(*iter);
                                    if(pObject->getLocation() == Coord(i,j)) {
                                        pObject->destroy();
                                    }
                                }
                            }
                        }
                    }


                    // destroy captured structure ...
                    delete pCapturedStructure;

                    // ... and create a new one
                    StructureBase* pNewStructure = owner->placeStructure(NONE, targetID, posX, posY, true);

                    pNewStructure->setOriginalHouse(origHouse);
                    pNewStructure->setHealth(oldHealth);
                    if(isSelected == true) {
                        pNewStructure->setSelected(true);
                        currentGame->getSelectedList().insert(pNewStructure->getObjectID());
                        currentGame->selectionChanged();
                        currentGame->getGameInterface().UpdateObjectInterface();
                    }

                    if(isSelectedByOtherPlayer == true) {
                        pNewStructure->setSelectedByOtherPlayer(true);
                        currentGame->getSelectedByOtherPlayerList().insert(pNewStructure->getObjectID());
                    }

                    if(containedUnitID != NONE) {
                        UnitBase* pNewUnit = owner->createUnit(containedUnitID);

                        pNewUnit->setRespondable(false);
                        pNewUnit->setActive(false);
                        pNewUnit->setVisible(VIS_ALL, false);
                        pNewUnit->setHealth(containedUnitHealth);

                        if(pNewUnit->getItemID() == Unit_Harvester) {
                            dynamic_cast<Harvester*>(pNewUnit)->setAmountOfSpice(containedHarvesterSpice);
                        }

                        if(pNewStructure->getItemID() == Structure_Refinery) {
                            Refinery* pRefinery = dynamic_cast<Refinery*>(pNewStructure);
                            pRefinery->book();
                            pRefinery->assignHarvester(dynamic_cast<Harvester*>(pNewUnit));
                        } else if(pNewStructure->getItemID() == Structure_RepairYard) {
                            RepairYard* pRepairYard = dynamic_cast<RepairYard*>(pNewStructure);
                            pRepairYard->book();
                            pRepairYard->assignUnit(pNewUnit);
                        }
                    }

                    // steal credits
                    pOwner->takeCredits(capturedSpice);
                    owner->addCredits(capturedSpice, false);
                    owner->checkSelectionLists();

                } else {
                    pCapturedStructure->handleDamage(50, NONE, getOwner());
                }
                // destroy unit indirectly
                setTarget(NULL);
                setHealth(0.0f);
                return;
            }
        } else if(target.getObjPointer() != NULL && target.getObjPointer()->isAStructure())	{
            Coord	closestPoint;
            closestPoint = target.getObjPointer()->getClosestPoint(location);

            if(blockDistance(location, closestPoint) <= 0.5f) {
                // destroy unit indirectly
                setTarget(NULL);
                setHealth(0.0f);
                return;
            }
        }
	}
}

void InfantryBase::destroy()
{
    if(currentGameMap->cellExists(location) && isVisible()) {
        Tile* pTile = currentGameMap->getCell(location);

        if(pTile->hasANonInfantryGroundObject() == true) {
            if(pTile->getNonInfantryGroundObject()->isAUnit()) {
                // squashed
                pTile->assignDeadUnit( currentGame->RandomGen.randBool() ? DeadUnit_Infantry_Squashed1 : DeadUnit_Infantry_Squashed2,
                                            owner->getHouseID(),
                                            Coord(lround(realX), lround(realY)) );

                if(isVisible(getOwner()->getTeam())) {
                    soundPlayer->playSoundAt(Sound_Squashed,location);
                }
            } else {
                // this unit has captured a building
            }

        } else if(getItemID() != Unit_Saboteur) {
            // "normal" dead
            pTile->assignDeadUnit( DeadUnit_Infantry,
                                        owner->getHouseID(),
                                        Coord(lround(realX), lround(realY)));

            if(isVisible(getOwner()->getTeam()))
                soundPlayer->playSoundAt((Sound_enum) getRandomOf(5,Sound_Scream1,Sound_Scream2,Sound_Scream3,Sound_Scream4,Sound_Scream5),location);
        }
    }

	GroundUnit::destroy();
}

void InfantryBase::move()
{
	if(!moving && currentGame->RandomGen.rand(0,40) == 0) {
		currentGameMap->viewMap(owner->getTeam(), location, getViewRange());
	}

	if(moving) {
		Coord	wantedReal;

		realX += xSpeed;
		realY += ySpeed;

		wantedReal.x = nextSpot.x*TILESIZE + TILESIZE/2,
		wantedReal.y = nextSpot.y*TILESIZE + TILESIZE/2;

		switch (cellPosition) {

		case 1:
			wantedReal.x -= TILESIZE/4;
			wantedReal.y -= TILESIZE/4;
			break;

		case 2:
			wantedReal.x += TILESIZE/4;
			wantedReal.y -= TILESIZE/4;
			break;

		case 3:
			wantedReal.x -= TILESIZE/4;
			wantedReal.y += TILESIZE/4;
			break;

		case 4:
			wantedReal.x += TILESIZE/4;
			wantedReal.y += TILESIZE/4;
			break;

		default:
			break;
		}

        // check if vehicle is on the first half of the way
		if(location != nextSpot) {
		    // check if vehicle is half way out of old cell
            if ((abs(location.x*TILESIZE - lround(realX) + TILESIZE/2) >= TILESIZE/2)
                || (abs(location.y*TILESIZE - lround(realY) + TILESIZE/2) >= TILESIZE/2))
            {
                unassignFromMap(location);	//let something else go in
                oldLocation = location;
                location = nextSpot;
		    }
		}

		if((std::abs((float)wantedReal.x - realX) <= 0.8f) && (std::abs((float)wantedReal.y - realY) <= 0.8f)) {
			realX = wantedReal.x;
			realY = wantedReal.y;

            if(forced && (location == destination) && !target) {
				setForced(false);
			}

			moving = false;
			justStoppedMoving = true;

			currentGameMap->viewMap(owner->getTeam(), location, getViewRange());

		}
	} else {
		justStoppedMoving = false;
	}

	checkPos();
}


void InfantryBase::setLocation(int xPos, int yPos)
{
	if (currentGameMap->cellExists(xPos, yPos) || ((xPos == INVALID_POS) && (yPos == INVALID_POS)))
	{
		oldCellPosition = cellPosition = INVALID_POS;
		GroundUnit::setLocation(xPos, yPos);
		switch (cellPosition)
		{
		case 1:
			realX -= TILESIZE/4;
			realY -= TILESIZE/4;
			break;
		case 2:
			realX += TILESIZE/4;
			realY -= TILESIZE/4;
			break;
		case 3:
			realX -= TILESIZE/4;
			realY += TILESIZE/4;
			break;
		case 4:
			realX += TILESIZE/4;
			realY += TILESIZE/4;
			break;
		default:
			break;
		}
	}
}

void InfantryBase::setSpeeds()
{
	if (oldCellPosition == INVALID_POS)
		fprintf(stderr, "InfantryBase::setSpeeds(): Infantry cell position  == INVALID_POS.\n");
	else if (cellPosition == oldCellPosition)	//havent changed infantry position
		GroundUnit::setSpeeds();
	else
	{
		int sx = 0, sy = 0, dx = 0, dy = 0;
		switch (oldCellPosition)
		{
		case 0: sx = sy = 0; break;
		case 1: sx = -TILESIZE/4; sy = -TILESIZE/4; break;
		case 2: sx = TILESIZE/4; sy = -TILESIZE/4; break;
		case 3: sx = -TILESIZE/4; sy = TILESIZE/4; break;
		case 4: sx = TILESIZE/4; sy = TILESIZE/4; break;
		}

		switch (cellPosition)
		{
		case 0: dx = dy = 0; break;
		case 1: dx = -TILESIZE/4; dy = -TILESIZE/4; break;
		case 2: dx = TILESIZE/4; dy = -TILESIZE/4; break;
		case 3: dx = -TILESIZE/4; dy = TILESIZE/4; break;
		case 4: dx = TILESIZE/4; dy = TILESIZE/4; break;
		}

		switch (drawnAngle)
		{
		case (RIGHT): dx += TILESIZE; break;
		case (RIGHTUP): dx += TILESIZE; dy -= TILESIZE; break;
		case (UP): dy -= TILESIZE; break;
		case (LEFTUP): dx -= TILESIZE; dy -= TILESIZE; break;
		case (LEFT): dx -= TILESIZE; break;
		case (LEFTDOWN): dx -= TILESIZE; dy += TILESIZE; break;
		case (DOWN): dy += TILESIZE; break;
		case (RIGHTDOWN): dx += TILESIZE; dy += TILESIZE; break;
		}

		dx -= sx;
		dy -= sy;

		float scale = currentGame->objectData.data[itemID].maxspeed/std::sqrt((float)(dx*dx + dy*dy));
		xSpeed = dx*scale;
		ySpeed = dy*scale;
	}
}

void InfantryBase::squash()
{
	destroy();
	return;
}

void InfantryBase::playConfirmSound() {
	soundPlayer->playSound((Sound_enum) getRandomOf(2,MovingOut,InfantryOut));
}

void InfantryBase::playSelectSound() {
	soundPlayer->playSound(YesSir);
}
